/**
 * @file common.c
 *
 * Copyright (C) IBM Corporation 2006
 */

#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/uio.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include "dr.h"
#include "ofdt.h"

/**
 * set_debug
 * @brief Common routine to set the debuig level and possibly librtas debug
 *
 * @param level debug level to set debug to
 */
void
set_debug(int level)
{
	debug = level;

	if (debug >= 4)
		debug_fd = open("/var/log/drmgr", O_RDWR | O_CREAT | O_TRUNC);

	if (debug >= 14)
		rtas_set_debug(debug);
}

void
clear_debug()
{
	if (debug_fd)
		close(debug_fd);

	if (debug >= 14)
		rtas_set_debug(0);
}
/**
 * dr_lock
 * @brief Attempt to lock a token
 *
 * This will attempt to lock a token (either file or directory) and wait
 * a specified amount of time if the lock cannot be granted.
 *
 * @param token token to lock
 * @param timeout timeout to wait before returning
 * @returns lock id if successful, -1 otherwise
 */
int
dr_lock(int timeout)
{
	struct flock    dr_lock_info;
	int             dr_lock_fd;
	int             rc;
	mode_t          old_mode;
	long            start_time;	/* Beginning time. Used for timeout */
	long            current_time;
	int             first_try;

	old_mode = umask(0);
	dr_lock_fd = open(DR_LOCK_FILE, O_RDWR | O_CREAT,
			  S_IRUSR | S_IRGRP | S_IROTH);
	if (dr_lock_fd < 0)
		return -1;

	umask(old_mode);
	dr_lock_info.l_type = F_WRLCK;
	dr_lock_info.l_whence = SEEK_SET;
	dr_lock_info.l_start = 0;
	dr_lock_info.l_len = 0;

	first_try = 1;
	start_time = time((long *) 0);	/* Pass 0 so that current time is
					 * returned */
	while (first_try || timeout == DR_WAIT ||
	       timeout > (current_time = time((long *) 0)) - start_time)
	{
		if (!first_try)
			sleep(1);
		first_try = 0;

		rc = fcntl(dr_lock_fd, F_SETLK, &dr_lock_info);
		if (rc != -1)
			return (dr_lock_fd);

		if (rc == -1 && errno == EACCES)
			continue;

	}

	close(dr_lock_fd);
	perror(DR_LOCK_FILE);
	return (-1);
}

/**
 * dr_unlock
 * @brief unlock a lock granted via dr_lock()
 *
 * @param lock_id
 * @returns 0 on success, -1 otherwise
 */
int
dr_unlock(int lock_id)
{
	struct flock    dr_lock_info;

	dr_lock_info.l_whence = SEEK_SET;
	dr_lock_info.l_start = 0;
	dr_lock_info.l_len = 0;
	dr_lock_info.l_type = F_UNLCK;
	if (fcntl(lock_id, F_SETLK, &dr_lock_info) < 0)
		return (-1);

	close(lock_id);
	return (0);

}

/**
 * add_node
 * @brief Add the specified node(s) to the device tree
 *
 * This will add the specified node(s) to the /proc Open Firmware
 * tree and the kernel Open Firmware tree.
 *
 * @param add_path
 * @param new_nodes
 * @returns 0 on success
 */
int
add_node(char *add_path, struct of_node *new_nodes)
{
	int rc = 0;
	int nprops = 0;
	int fd;
	struct of_property *prop;
	char *buf, *pos;
	size_t bufsize = 0;

	dbg("Adding node at %s to kernel\n", add_path);

	/* determine the total size of the buffer we need to allocate */
	bufsize += strlen("add_node ");
	bufsize += strlen(add_path) + 1;
	for (prop = new_nodes->properties; prop; prop = prop->next) {
		char tmp[16] = { '\0' }; /* for length */

		nprops++;
		bufsize += strlen(prop->name) + 1; /* name + space */
		sprintf(tmp, "%d", prop->length);

		bufsize += strlen(tmp) + 1;	/* length + space*/
		bufsize += prop->length + 1;	/* value + space */
	}

	if (! nprops) {
		err_msg("new_nodes have no properties\n");
		return -1;
	}

	buf = malloc(bufsize);
	if (buf == NULL) {
		err_msg("Failed to allocate buffer to write to kernel\n");
		return errno;
	}

	memset(buf, '\0', bufsize);

	strcpy(buf, "add_node ");
	strcat(buf, add_path);
	strcat(buf, " ");
	pos = buf + strlen(buf);

	/* this is less than optimal, iterating over the entire buffer
	 * for every strcat...
	 */
	for (prop = new_nodes->properties; prop; prop = prop->next) {
		char tmp[16] = { '\0' }; /* for length */

		memcpy(pos, prop->name, strlen(prop->name));
		pos += strlen(prop->name);
		*pos++ = ' ';

		sprintf(tmp, "%d", prop->length);
		memcpy(pos, tmp, strlen(tmp));
		pos += strlen(tmp);
		*pos++ = ' ';

		memcpy(pos, prop->value, prop->length);
		pos += prop->length;
		*pos++ = ' ';
	}
	*pos = '\0';

	/* dump the buffer to a file in /tmp for debugging */
	if (debug >= 4) {
		int dumpfd;

		dbg("Dumping command buffer to /tmp/ofdt.cmd\n");

		dumpfd = creat("/tmp/ofdt.cmd", 0644);
		write(dumpfd, buf, bufsize);
		close(dumpfd);
	}

	fd = open(OFDTPATH, O_WRONLY);
	if (fd <= 0) {
		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
		return errno;
	}

	rc = write(fd, buf, bufsize);
	if (rc <= 0)
		err_msg("Write to %s failed: %s\n", OFDTPATH, strerror(errno));
	else
		rc = 0;

	close(fd);
	return rc;
}

/**
 * process_add_nodes
 * @brief Add new nodes to the device tree.
 *
 * @param new_nodes
 * @returns 0 on success, !0 otherwise
 */
int
process_add_nodes(struct of_node *new_nodes)
{
	char add_path[DR_PATH_MAX];
	int rc;

	/* When adding CPU's the preceding /cpus is not included in the
	 * node_name of the node passed in.  Since this is the only type
	 * of node that is missing part of the path we check for it here
	 */
	if (strncmp(new_nodes->name, "PowerPC,", 8) == 0)
		sprintf(add_path, "/cpus/%s", new_nodes->name);
	else
		sprintf(add_path, "/%s", new_nodes->name);

	rc = add_node(add_path, new_nodes);
	if (rc)
		return rc;

	if (new_nodes->child)
		process_add_nodes(new_nodes->child);

	if (new_nodes->sibling)
		process_add_nodes(new_nodes->sibling);

	return rc;
}

/**
 * remove_node
 * @brief Remove the specified node to the device tree
 *
 * This will remove the specified node to the /proc Open Firmware
 * tree and the kernel Open Firmware tree.
 *
 * @param cmd
 * @returns 0 on success
 */
int
remove_node(char *path)
{
	int rc = 0;
	int fd;
	int cmdlen;
	char buf[DR_PATH_MAX];

	dbg("Removing node at %s from kernel\n", path);

	memset(buf, 0, DR_PATH_MAX);
	sprintf(buf, "remove_node %s", path);

	cmdlen = strlen(buf);

	fd = open(OFDTPATH, O_WRONLY);
	if (fd <= 0) {
		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
		return errno;
	}

	rc = write(fd, buf, cmdlen);
	if (rc != cmdlen)
		err_msg("Write to %s failed: %s\n", OFDTPATH, strerror(errno));
	else
		rc = 0;

	close(fd);
	return rc;
}

/**
 * add_nodes_to_kernel
 *
 * Process new_nodes from configure_connector and call add_node to
 * add new nodes to /proc device-tree and the kernel's device tree.
 *
 * @param root_path
 * @param new_nodes
 * @returns 0 on success, !0 otherwise
 */
int
add_nodes_to_kernel(char *root_path, struct of_node *new_nodes)
{
	struct of_node *node;
	char add_path[DR_PATH_MAX];
	int rc = 0;

	for (node = new_nodes; node; node = node->sibling) {
		if (!strcmp(root_path, "/"))
			sprintf(add_path, "/%s", node->name);
		else if (*root_path == '/')
			sprintf(add_path, "%s/%s", root_path, node->name);
		else
			sprintf(add_path, "/%s/%s", root_path, node->name);


		rc = add_node(add_path, node);
		if (rc)
			return rc;

		if (node->child) {
			rc = add_nodes_to_kernel(add_path, node->child);
			if (rc)
				return rc;
		}
	}

	return rc;
}

/**
 * remove_nodes_from_kernel
 *
 * Remove all device nodes and children device nodes from Open Firmware
 * device tree
 *
 * @param root_path
 * @returns 0 on success, !0 otherwise
 */
int
remove_nodes_from_kernel(char *root_path)
{
        DIR *d;
        struct dirent *de;
        struct stat sb;
	char dir_name[DR_PATH_MAX];
	int rc;

	sprintf(dir_name, "/proc/device-tree%s", root_path);
        if ((lstat(dir_name, &sb) == 0)
	    && (S_ISDIR(sb.st_mode)) && (! S_ISLNK(sb.st_mode))) {
                d = opendir(dir_name);
                if (d == NULL) {
			err_msg("Could not open %s: %s\n", dir_name,
				strerror(errno));
                        return 1;
                }

                /* make sure d has no sudir */
                while ((de = readdir(d)) != NULL) {
                        char subdir_name[DR_PATH_MAX];
                        struct stat buf;

                        if ((! strcmp(de->d_name, "."))
			    || (! strcmp(de->d_name, "..")))
                                continue;

                        sprintf(subdir_name, "%s/%s", dir_name, de->d_name);
                        if ((lstat(subdir_name, &buf) == 0)
			    && (S_ISDIR(buf.st_mode))
			    && (! S_ISLNK(buf.st_mode))) {
				char sub_path[DR_PATH_MAX];
				sprintf(sub_path, "%s/%s", root_path,
					de->d_name);
                                remove_nodes_from_kernel(sub_path);
				rewinddir(d);
			}
                }

                closedir(d);
		rc = remove_node(root_path);
        }

        return rc;
}

/**
 * get_att_prop
 * @brief find the value for a given property/attribute.
 *
 * @param path path to the property/attribute to retrieve
 * @param name name of the property/attribute to retrieve
 * @param buf buffer to write property/attribute to
 * @param buf_sz size of the buffer
 * @returns 0 on success, -1 otherwise
 */
static int
get_att_prop(const char *path, const char *name, char *buf, size_t buf_sz,
	     const char *attr_type)
{
	FILE *fp;
	int rc;
	char dir[DR_PATH_MAX];

	if (buf == NULL)
		return -1;

	if (name != NULL)
		sprintf(dir, "%s/%s", path, name);
	else
		sprintf(dir, "%s", path);

	fp = fopen(dir, "r");
	if (fp == NULL)
		return -1;

	memset(buf, 0, buf_sz);

	/* Yes, this is sort of a hack but we only read properties from
	 * either /proc or sysfs so it works and is cheaper than a strcmp()
	 */
	switch (dir[1]) {
	    case 'p':	/* /proc */
		rc = fread(buf, buf_sz, 1, fp);
		break;

	    case 's':	/* sysfs */
		rc = fscanf(fp, attr_type, (int *)buf);
		break;
	}

	fclose(fp);
	return 0;
}

/**
 * get_property
 * @brief retrieve a device-tree property from /proc
 *
 * @param path path to the property to retrieve
 * @param name name of the property to retrieve
 * @param buf buffer to write property to
 * @param buf_sz size of the buffer
 * @returns 0 on success, !0 otherwise
 */
int
get_property(const char *path, const char *property, void *buf, size_t buf_sz)
{
	return get_att_prop(path, property, buf, buf_sz, NULL);
}

/**
 * get_int_attribute
 * @brief retrieve an integer device attribute from sysfs
 *
 * @param path path to the attribute to retrieve
 * @param name name of the attribute to retrieve
 * @param buf buffer to write attribute to
 * @param buf_sz size of the buffer
 * @returns 0 on success, -1 otherwise
 */
int
get_int_attribute(const char *path, const char *attribute, void *buf,
		  size_t buf_sz)
{
	return get_att_prop(path, attribute, buf, buf_sz, "%i");
}

/**
 * get_str_attribute
 * @brief retrieve a string device attribute from sysfs
 *
 * @param path path to the attribute to retrieve
 * @param name name of the attribute to retrieve
 * @param buf buffer to write attribute to
 * @param buf_sz size of the buffer
 * @returns 0 on success, -1 otherwise
 */
int
get_str_attribute(const char *path, const char *attribute, void *buf,
		  size_t buf_sz)
{
	return get_att_prop(path, attribute, buf, buf_sz, "%s");
}

/**
 * get_property_size
 * @brief retrieve the size of a property
 *
 * @param path path to the property to retrieve
 * @param name name of the property to retrieve
 * @returns size of the property
 */
int
get_property_size(const char *path, const char *property)
{
	char dir[DR_PATH_MAX];
	struct stat sb;

	if (property != NULL)
		sprintf(dir, "%s/%s", path, property);
	else
		sprintf(dir, "%s", path);

	stat(dir, &sb);
	return sb.st_size;
}

/**
 * signal_handler
 * @brief Dummy signal handler
 *
 * A dummy signal handler for registering for the SIGDANGER
 * signal so that this command won't be killed when paging space
 * gets low.  The SIGDANGER signal is masked though, so this handler
 * will never actually get called.
 *
 * @param signal
 * @param code
 * @param scp
 * @retuen 0 always
 */
int
signal_handler(int signal, int code, struct sigcontext * scp)
{
	return 0;
}

/**
 * mask_signals
 * @brief Mask signals so that dynamic reconfig operations won't be interrupted
 *
 * @returns 0 on success, !0 otherwise
 */
int
mask_signals()
{
	struct sigaction action;/* parameters of sigaction */
	struct sigaction oldaction;
	sigset_t        sigset;
	int             rc;

	/* Start by defining a signal handler for SIGDANGER */
	memset(&action, '\0', sizeof(struct sigaction));
	action.sa_handler = (void (*) (int)) signal_handler;

	/* sigaction (SIGDANGER, &action, &oldaction); */
	sigaction(SIGPWR, &action, &oldaction);	/* power fail */

	/* Now set up a mask with all signals masked */
	sigfillset(&sigset);

	/* Clear mask bits for signals we don't want to mask */
	sigdelset(&sigset, SIGBUS);
	sigdelset(&sigset, SIGXFSZ);
	sigdelset(&sigset, SIGSEGV);
	sigdelset(&sigset, SIGTRAP);
	sigdelset(&sigset, SIGILL);
	sigdelset(&sigset, SIGFPE);
	sigdelset(&sigset, SIGSYS);
	sigdelset(&sigset, SIGPIPE);
	sigdelset(&sigset, SIGVTALRM);
	sigdelset(&sigset, SIGALRM);

	/* Now block all remaining signals */
	rc = sigprocmask(SIG_BLOCK, &sigset, NULL);
	if (rc)
		return -1;

	return 0;
}

char *php_slot_type_msg[]={
	"",
	"PCI 32 bit, 33MHz, 5 volt slot",
	"PCI 32 bit, 50MHz, 5 volt slot",
	"PCI 32 bit, 33MHz, 3.3 volt slot",
	"PCI 64 bit, 33MHz, 5 volt slot",
	"PCI 64 bit, 50MHz, 5 volt slot",	/* 5 */
	"PCI 64 bit, 33MHz, 3.3 volt slot",
	"PCI 32 bit, 66MHz, 3.3 volt slot",
	"PCI 64 bit, 66MHz, 3.3 volt slot",
	"",				/* we don't have connector types for */
	"",				/* 9 or 10, so skip these indices.   */
	"PCI-X capable, 32 bit, 66MHz slot",
	"PCI-X capable, 32 bit, 100MHz slot",
	"PCI-X capable, 32 bit, 133MHz slot",
	"PCI-X capable, 64 bit, 66MHz slot",
	"PCI-X capable, 64 bit, 100MHz slot",	/* 15 */
	"PCI-X capable, 64 bit, 133MHz slot",
	"PCI-X capable, 64 bit, 266MHz slot",
	"PCI-X capable, 64 bit, 533MHz slot",
	"PCI-E capable, Rev 1, 1x lanes",
	"PCI-E capable, Rev 1, 2x lanes", 	/* 20 */
	"PCI-E capable, Rev 1, 4x lanes",
	"PCI-E capable, Rev 1, 8x lanes",
	"PCI-E capable, Rev 1, 16x lanes",
	"PCI-E capable, Rev 1, 32x lanes",
	"PCI-E capable, Rev 2, 1x lanes",	/* 25 */
	"PCI-E capable, Rev 2, 2x lanes",
	"PCI-E capable, Rev 2, 4x lanes",
	"PCI-E capable, Rev 2, 8x lanes",
	"PCI-E capable, Rev 2, 16x lanes",
	"PCI-E capable, Rev 2, 32x lanes",	/* 30 */
};

char *
node_type(struct node *node)
{
	int desc_msg_num;
	char *desc;

	desc_msg_num = atoi(node->drc_type);
	if ((desc_msg_num >= 1 &&  desc_msg_num <= 8) ||
	    (desc_msg_num >= 11 && desc_msg_num <= 30))
		desc = php_slot_type_msg[desc_msg_num];
	else {
		switch (node->dev_type) {
			case PCI_DLPAR_DEV:
				desc = "Logical I/O Slot";
				break;
			case VIO_DEV:
				desc = "Virtual I/O Slot";
				break;
			case HEA_DEV:
				desc = "HEA I/O Slot";
				break;
			case HEA_PORT_DEV:
				desc = "HEA Port I/O Slot";
				break;
			default:
				desc = "Unknown slot type";
				break;
		}
	}

	return desc;
}

/**
 * kernel_check
 * @brief Validate that the specified file exists for this kernel
 *
 * @param path full path to the file to validate kernels ability
 * @returns 0 on success, !0 otherwise
 */
int
kernel_check(const char *path)
{
	struct stat sbuf;
	int rc;

	dbg("Validating kernel file: %s\n", path);
	
	rc = stat(path, &sbuf);
	return rc;
}

/**
 * valid_platform
 * @brief Validate that the platfomr we are on matches the one specified
 *
 * @param platform Name of platform to validate
 * @return 1 if valid, 0 otherwise
 */
int
valid_platform(const char *platform)
{
	char buf[128];
	int rc;

	rc = get_property(OFDT_BASE, "device_type", buf, 128);
	if (rc) {
		err_msg("Cannot open %s: %s\n", PLATFORMPATH, strerror(errno));
		return 0;
	}

	if (strcmp(buf, platform)) {
		err_msg("This command is not supported for %s platforms.\n",
			platform);
		return 0;
	}

	return 1;
}
