/**
 * @file common_cpu.c
 * @brief Common routines for cpu data
 *
 * Copyright (C) IBM Corporation 2006
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <librtas.h>
#include "dr.h"
#include "drcpu.h"
#include "ofdt.h"

/* format strings for easy access */
#define DR_THREAD_DIR_PATH "/sys/devices/system/cpu/cpu%d"
#define DR_THREAD_ONLINE_PATH "/sys/devices/system/cpu/cpu%d/online"
#define DR_THREAD_PHYSID_PATH "/sys/devices/system/cpu/cpu%d/physical_id"

#define DR_CPU_INTSERVERS_PATH \
        "/proc/device-tree/cpus/%s/ibm,ppc-interrupt-server#s"

/**
 * id_to_thread
 * @brief Find the thread struct for a given id
 *
 * @param id id of thread to find
 * @returns pointer to thread struct on success, NULL otherwise
 */
struct thread_info *
id_to_thread(int id, struct dr_info *dr_info)
{
	struct thread_info *thread;

	for (thread = dr_info->all_threads; thread; thread = thread->next) {
		if (thread->id == id)
			return thread;
	}

	return NULL;
}

/**
 * cpu_id
 * @brief Get a cpu's physical id
 *
 * @param cpu
 * @returns cpu id on success, -1 otherwise
 */
static int
cpu_id(struct cpu_info *cpu)
{
	int id, rc;

	rc = get_property(cpu->path, "reg", &id, sizeof(id));
	return rc ? rc : id;
}

/**
 * cpu_nr_threads
 * @brief Get the number of threads this cpu has
 *
 * @param cpu cpu to count threads on
 */
static int
cpu_nr_threads(struct cpu_info *cpu)
{
	struct stat sb;
	char path[DR_PATH_MAX];

	/* The number of threads is the number of 32-bit ints in the
	 * cpu's ibm,ppc-interrupt-server#s property.
	 */
	sprintf(path, "/proc/device-tree/cpus/%s/ibm,ppc-interrupt-server#s",
		cpu->name);

	if (stat(path, &sb))
		/* Assume no SMT */
		return 1;

	return sb.st_size / 4;
}

/**
 * cpu_drc_index
 * @brief Get a cpu's drc-index
 *
 * @param cpu cpu_info to get drc index for
 * @returns drc index on success, -1 otherwise
 */
int
cpu_drc_index(struct cpu_info *cpu)
{
	int rc, index;
	rc = get_property(cpu->path, "ibm,my-drc-index", &index, sizeof(index));

	return rc ? rc : index;
}

/**
 * cpu_drc_name
 * @brief REtrieve the name of a drc for the cpu
 *
 * @param cpu cpu to get drc name for
 * @returns drc name on success, NULL otherwise
 */
char *
cpu_drc_name(struct cpu_info *cpu, struct dr_info *dr_info)
{
	return drc_index_to_name(cpu_drc_index(cpu), dr_info->all_drcs);
}

/**
 * cpu_owns_physid
 * @brief Check a cpu property for an id
 *
 * Check this cpu's ibm,ppc-interrupt-server#s property
 * to see whether id occurs in it; if so, the id belongs
 * to the cpu.  Return 1 if the id is found, 0 if not
 *
 * @param cpu
 * @param id
 * @returns 0 on success, !0 otherwise
 */
static int
cpu_owns_physid(struct cpu_info *cpu, int id)
{
	FILE *f;
	char path[DR_PATH_MAX];
	int server_id, found = 0;

	sprintf(path, DR_CPU_INTSERVERS_PATH, cpu->name);

	f = fopen(path, "r");
	if (f == NULL) {
		/* fall back on checking the reg property for the id */
		if (cpu_id(cpu) == id)
			return 1;
		return 0;
	}

	while ((fread(&server_id, sizeof(server_id), 1, f))) {
		if (server_id == id) {
			found = 1;
			break;
		}
	}

	fclose(f);
	return found;
}

/**
 * physid_to_cpu
 *
 * @param id
 * returns pointer cpu_info on success, NULL otherwise
 */
static struct cpu_info *
physid_to_cpu(int id, struct dr_info *dr_info)
{
	struct cpu_info *cpu;

	if (id == -1)
		return NULL;

	for (cpu = dr_info->all_cpus; cpu; cpu = cpu->next) {
		if (cpu_owns_physid(cpu, id))
			return cpu;
	}

	return NULL;
}

/**
 * thread_physid
 *
 * @param thread
 */
static int
thread_physid(struct thread_info *thread)
{
	int rc, phys_id;

	/* there may be needed some sort of check to ensure that we don't
	 * return 0 for threads which don't actually belong to this lpar
	 * i.e. check dlpar_state
	 */
	rc = get_int_attribute(thread->path, "physical_id", &phys_id,
			       sizeof(phys_id));

	return rc ? rc : phys_id;
}

/**
 * thread_to_cpu
 * @brief Retrieve cpu_info for a given thread
 *
 * @param thread
 * @returns pointer to cpu_info on success, NULL otherwise
 */
struct cpu_info *
thread_to_cpu(struct thread_info *thread, struct dr_info *dr_info)
{
	return physid_to_cpu(thread_physid(thread), dr_info);
}

/**
 * free_thread_info
 * @brief free the thread list and any associated allocated memory
 *
 * @param thread_list list of threads to free
 */
void
free_thread_info(struct thread_info *thread_list)
{
	free(thread_list);
}

/**
 * init_thread_info
 * @brief Initialize thread data
 *
 * Initialize global thread or "logical cpu" data. The physical cpu data
 * initialization must have been completed before this is called.
 *
 * @returns pointer to thread_info on success, NULL otherwise
 */
static struct thread_info *
init_thread_info(void)
{
	struct thread_info *thread_list = NULL;
	int i = 0;
	struct stat s;
	char path[DR_PATH_MAX];
	int nr_threads;

	if (stat("/sys/devices/system/cpu", &s)) {
		err_msg("%s\n", strerror(errno));
		return NULL;
	}

	nr_threads = s.st_nlink - 2;

	dbg("Expecting %d threads\n", nr_threads);

	thread_list = malloc(sizeof(*thread_list) * nr_threads);
	if (thread_list == NULL) {
		err_msg("%s\n", strerror(errno));
		return NULL;
	}

	/* Assume the range of possible thread id's is contiguous starting
	 * from 0.  The threads must be in order, hence the array treatment
	 * here.  The rest of the code in this file simply uses the
	 * linked list.
	 */
	sprintf(path, DR_THREAD_DIR_PATH, i);
	for (i = 0; 0 == stat(path, &s); i++) {
		struct thread_info *new_thread = &thread_list[i];

		dbg("Found thread %s\n", path);

		new_thread->id = i;
		sprintf(new_thread->path, path);

		if (new_thread->id == nr_threads - 1)
			/* this is the last thread */
			new_thread->next = NULL;
		else
			new_thread->next = &thread_list[i + 1];

		sprintf(path, DR_THREAD_DIR_PATH, i + 1);
	}

	return thread_list;
}

/**
 * free_cpu_info
 * @brief free the list of cpu structs and any assocaiated allocated memory
 *
 * @param cpu_list list of cpus to free
 */
void
free_cpu_info(struct cpu_info *cpu_list)
{
	struct cpu_info *tmp = cpu_list;

	while (tmp != NULL) {
		cpu_list = cpu_list->next;
		free(tmp);
		tmp = cpu_list;
	}
}

/**
 * init_cpu_info
 * @brief Initialize cpu information
 *
 * @returns pointer to cpu_info on success, NULL otherwise
 */
static struct cpu_info *
init_cpu_info(struct dr_info *dr_info)
{
	struct cpu_info *cpu_list = NULL;
	DIR *d;
	struct dirent *ent;

	d = opendir("/proc/device-tree/cpus");
	if (d == NULL) {
		err_msg("%s\n", strerror(errno));
		return NULL;
	}

	while ((ent = readdir(d))) {
		char path[DR_PATH_MAX];
		struct stat sb;

		/* skip everything but directories */
		sprintf(path, "/proc/device-tree/cpus/%s", ent->d_name);

		if (lstat(path, &sb)) {
			err_msg("%s\n", strerror(errno));
			return NULL;
		}

		if (!S_ISDIR(sb.st_mode))
			continue;

		if (strstr(ent->d_name, "PowerPC")) {
			struct cpu_info *cpu;

			cpu = malloc(sizeof(*cpu));
			if (cpu == NULL) {
				err_msg("Could not allocate cpu info.\n%s\n",
					strerror(errno));
				free_cpu_info(cpu_list);
				cpu_list = NULL;
				break;
			}

			memset(cpu, -1, sizeof(cpu));

			sprintf(cpu->name, ent->d_name);
			sprintf(cpu->path, path);

			cpu->next = cpu_list;
			cpu_list = cpu;

			dbg("Found cpu %s\n", cpu->name);
		}
	}

	closedir(d);

	return cpu_list;
}

/**
 * drc_index_to_cpu
 * @brief Convert a drc index to cpu information
 *
 * @param index drc index
 * @returns pointer to cpu_info on success, NULL otherwise
 */
struct cpu_info *
drc_index_to_cpu(int index, struct dr_info *dr_info)
{
	struct cpu_info *cpu;

	for (cpu = dr_info->all_cpus; cpu != NULL; cpu = cpu->next) {
		if (cpu_drc_index(cpu) == index)
			return cpu;
	}

	return NULL;
}

/**
 * get_available_drc_index
 *
 * Find the first index which does not correspond to a cpu
 * already owned by the partition.
 *
 * @returns drc index on success, 0 otherwise
 */
static int
get_available_drc_index(struct dr_info *dr_info)
{
	struct dr_connector *drc;

	for (drc = dr_info->all_drcs; drc; drc = drc->next) {
		int rc;
		struct cpu_info *cpu;

		dbg("checking drc-index 0x%x ... ", drc->index);
		rc = dr_entity_sense(drc->index);
		if (rc != STATE_UNUSABLE) {
			dbg("unavailable\n");
			continue;
		}

		if ((cpu = drc_index_to_cpu(drc->index, dr_info))) {
			dbg("already owned by this LPAR (%s)\n", cpu->name);
			continue;
		}

		dbg("available\n");
		return (drc->index);
	}

	return 0;
}

/**
 * cache_phandle
 *
 * @param cache
 */
int
cache_phandle(struct cache_info *cache)
{
	int rc, phandle;

	rc = get_property(cache->path, "ibm,phandle", &phandle,
			  sizeof(phandle));

	return rc ? rc : phandle;
}

/**
 * cpu_l2cache
 *
 * @param cpu
 * @returns L2 cache on success, -1 otherwise
 */
static int
cpu_l2cache(struct cpu_info *cpu)
{
	int rc, l2cache;

	rc = get_property(cpu->path, "l2-cache", &l2cache, sizeof(l2cache));
	return rc ? rc : l2cache;
}

/**
 * cache_l2cache
 *
 * @param cache
 * @returns L2 cache on success, -1 otherwise
 */
static int cache_l2cache(struct cache_info *cache)
{
	int rc, l2cache;

	rc = get_property(cache->path, "l2-cache", &l2cache, sizeof(l2cache));
	return rc ? rc : l2cache;
}

/**
 * cpu_get_dependent_cache
 * @brief Return the cache whose ibm,phandle matches the given cpu's
 *        l2-cache property.
 *
 * @param cpu
 * @returns pointer to cache_info on success, NULL otherwise
 */
struct cache_info *
cpu_get_dependent_cache(struct cpu_info *cpu, struct dr_info *dr_info)
{
	int phandle;
	struct cache_info * cache;

	phandle = cpu_l2cache(cpu);

	for (cache = dr_info->all_caches; cache != NULL; cache = cache->next) {
		if (cache_phandle(cache) == phandle) {
			dbg("found %s for %s\n", cache->name, cpu->name);
			return cache;
		}
	}

	return NULL;
}

/**
 * cache_get_dependent_cache
 * @brief Return the cache whose ibm,phandle matches the given cache's
 *       l2-cache property.
 *
 * @param cache
 * @returns pointer to cache_info on success, NULL otherwise
 */
struct cache_info *
cache_get_dependent_cache(struct cache_info *cache, struct dr_info *dr_info)
{
	int phandle;
	struct cache_info *c;

	phandle = cache_l2cache(cache);

	for (c = dr_info->all_caches; c != NULL; c = c->next) {
		if (cache_phandle(c) == phandle) {
			dbg("found %s for %s\n", c->name, cache->name);
			return c;
		}
	}

	return NULL;
}

/**
 * cpu_remove_devnode
 * @brief Remove the given cpu device node from the kernel's device tree.
 *        Also remove the cpu from the global cpu list.
 *
 * @param cpu
 * @returns
 */
static int
cpu_remove_devnode(struct cpu_info *cpu, struct dr_info *dr_info)
{
	char path[DR_PATH_MAX];
	int rc;

	dbg("removing %s from device tree\n", cpu->name);

	sprintf(path, "/cpus/%s", cpu->name);

	if ((rc = remove_node(path)) == 0) {
		struct cpu_info *c = dr_info->all_cpus;

		if (c == cpu)
			dr_info->all_cpus = cpu->next;
		else {
			while (c->next != cpu)
				c = c->next;
			c->next = cpu->next;
		}

		free(cpu);
	}

	return rc;
}

/**
 * cache_remove_devnode
 * @brief Remove the given cache device node from the kernel's device tree.
 *        Also remove the cache from the global cache list.
 *
 * @param cache
 * @returns
 */
static int
cache_remove_devnode(struct cache_info *cache, struct dr_info *dr_info)
{
	char path[DR_PATH_MAX];
	int rc;

	dbg("removing %s from device tree\n", cache->name);

	sprintf(path, "/cpus/%s", cache->name);

	if ((rc = remove_node(path)) == 0) {
		struct cache_info *c = dr_info->all_caches;

		if (c == cache)
			dr_info->all_caches = cache->next;
		else {
			while (c->next != cache)
				c = c->next;
			c->next = cache->next;
		}

		free(cache);
	}

	return rc;
}

/**
 * cache_get_use_count
 * @brief Calculate a cache's use count
 *
 * @param cache
 * @returns
 */
static int
cache_get_use_count(struct cache_info *cache, struct dr_info *dr_info)
{
	int count = 0;
	struct cpu_info *cpu;
	struct cache_info *c;

	/* Any cpu or cache node whose 'l2-cache' property matches
	 * the ibm,phandle is using our cache
	 */
	for (cpu = dr_info->all_cpus; cpu != NULL; cpu = cpu->next) {
		if (cache == cpu_get_dependent_cache(cpu, dr_info)) {
			dbg("%s->ibm,phandle 0x%x == %s->l2-cache\n",
			    cache->name, cache_phandle(cache), cpu->name);
			count++;
		}
	}

	for (c = dr_info->all_caches; c != NULL; c = c->next) {
		if (cache == cache_get_dependent_cache(c, dr_info)) {
			dbg("%s->ibm,phandle 0x%x == %s->l2-cache\n",
			    cache->name, cache_phandle(cache), c->name);
			count++;
		}
	}

	return count;
}

/**
 * refresh_cpu_info
 *
 * @returns 0 on success, !0 otherwise
 */
static int
refresh_cpu_info(struct dr_info *dr_info)
{
	free_cpu_info(dr_info->all_cpus);
	dr_info->all_cpus = init_cpu_info(dr_info);

	if (dr_info->all_cpus == NULL) {
		err_msg("failed to refresh cpu info\n");
		return 1;
	}

	return 0;
}

/**
 * free_cache_info
 * @brief free the cache list and any associated allocated memory
 *
 * @param cache_list list of cache structs to free
 */
void
free_cache_info(struct cache_info *cache_list)
{
	struct cache_info *tmp = cache_list;

	while (tmp != NULL) {
		cache_list = cache_list->next;
		free(tmp);
		tmp = cache_list;
	}
}

/**
 * init_cache_info
 *
 * @returns pointer to cache_info on success, NULL otherwise
 */
static struct cache_info *
init_cache_info(struct dr_info *dr_info)
{
	struct cache_info *cache_list = NULL;
	DIR *d;
	struct dirent *ent;

	d = opendir("/proc/device-tree/cpus");
	if (d == NULL) {
		err_msg("%s\n", strerror(errno));
		return NULL;
	}

	while ((ent = readdir(d))) {
		char path[DR_PATH_MAX];
		struct stat sb;

		/* skip everything but directories */
		sprintf(path, "/proc/device-tree/cpus/%s", ent->d_name);
		if (lstat(path, &sb)) {
			err_msg("%s\n", strerror(errno));
			break;
		}

		if (!S_ISDIR(sb.st_mode))
			continue;

		if (strstr(ent->d_name, "-cache@")) {
			struct cache_info *cache;

			cache = malloc(sizeof(*cache));
			if (cache == NULL) {
				err_msg("Could not allocate cache info.\n%s\n",
					strerror(errno));
				free_cache_info(cache_list);
				cache_list = NULL;
				break;
			}

			sprintf(cache->name, ent->d_name);
			sprintf(cache->path, path);

			cache->next = cache_list;
			cache_list = cache;

			dbg("Found cache %s\n", cache->name);
		}
	}

	closedir(d);

	return cache_list;
}

/**
 * refresh_cache_info
 *
 * @returns 0 on success, !0 otherwise
 */
static int
refresh_cache_info(struct dr_info *dr_info)
{
	/* Some systems do not have cache device nodes.  Assume
	 * that if we did not find any cache nodes during initialization,
	 * we won't find any during a refresh.
	 */
	if (dr_info->all_caches == NULL)
		return 0;

	free_cache_info(dr_info->all_caches);
	dr_info->all_caches = init_cache_info(dr_info);

	if (dr_info->all_caches == NULL) {
		err_msg("failed to refresh info\n");
		return 1;
	}

	return 0;
}

/**
 * acquire_cpu
 *
 * Acquire a new cpu for this partition from the hypervisor.
 * If drc_index is zero, look up an available index to use.
 *
 * @param drc_index
 * @returns pointer to cpu_info on success, NULL otherwise
 */
struct cpu_info *
acquire_cpu(int drc_index, struct dr_info *dr_info)
{
	int rc;
	struct cpu_info *cpu;
	struct of_node *new_nodes; /* cpu and cache(s) */

	if (drc_index == 0
	    && 0 == (drc_index = get_available_drc_index(dr_info))) {
		err_msg("unable to find an available drc-index\n");
		return NULL;
	}

	rc = acquire_drc(drc_index, &new_nodes);
	if (rc)
		return NULL;

	/* Add the new nodes to the device tree */
	rc = process_add_nodes(new_nodes);
	refresh_cpu_info(dr_info);
	refresh_cache_info(dr_info);

	if (rc == 0) {
		/* Look up the new cpu */
		cpu = drc_index_to_cpu(drc_index, dr_info);
	}

	return cpu;
}

/**
 * release_cpu
 *
 * Release a cpu back to the hypervisor, using the cpu's drc-index.
 * The cpu must have already been offlined.  We remove the cpu from the
 * device tree only when the isolate and set_indicator have succeeded.
 * The caller must not use the value given for the cpu parameter after
 * this function has returned.
 *
 * @param cpu
 * @returns 0 on success, !0 otherwise
 */
int
release_cpu(struct cpu_info *cpu, struct dr_info *dr_info)
{
	int rc, drc_index = cpu_drc_index(cpu);
	struct cache_info *L2_cache, *L3_cache;

	/* Should we check to make sure all threads of the cpu are offline? */

	rc = release_drc(drc_index, 0);
	if (rc)
		return rc;

	/* remove any unused caches */
	L2_cache = cpu_get_dependent_cache(cpu, dr_info);

	rc = cpu_remove_devnode(cpu, dr_info);

	if (!rc && L2_cache && 0 == cache_get_use_count(L2_cache, dr_info)) {
		L3_cache = cache_get_dependent_cache(L2_cache, dr_info);
		if (0 == cache_remove_devnode(L2_cache, dr_info))
			if (L3_cache
			    && 0 == cache_get_use_count(L3_cache, dr_info))
				cache_remove_devnode(L3_cache, dr_info);
	}

	return rc;
}


/**
 * init_cpu_drc_info
 *
 * @returns pointer to drc_info on success, NULL otherwise
 */
int
init_cpu_drc_info(struct dr_info *dr_info)
{
	memset(dr_info, 0, sizeof(*dr_info));

	dr_info->all_cpus = init_cpu_info(dr_info);
	if (dr_info->all_cpus == NULL)
		return -1;

	dr_info->all_threads = init_thread_info();
	if (dr_info->all_threads == NULL)
		return -1;

	dr_info->all_drcs = init_drc_info("/cpus");
	if (dr_info->all_drcs == NULL)
		return -1;

	/* It is ok if no cache info is found; not all systems
	 * have cache device nodes.
	 */
	dr_info->all_caches = init_cache_info(dr_info);

	return 0;
}

/**
 * free_cpu_drc_info
 * @brief free the allocated lists hanging off of the drc_info struct
 *
 * @param dr_info dr_info struct to free memory from
 */
void
free_cpu_drc_info(struct dr_info *dr_info)
{
	free_cache_info(dr_info->all_caches);
	free_drc_info(dr_info->all_drcs);
	free_thread_info(dr_info->all_threads);
	free_cpu_info(dr_info->all_cpus);

	memset(dr_info, 0, sizeof(*dr_info));
}

/**
 * set thread_state
 * @brief Set the thread state
 *
 * Set the given thread to the desired state.  This just writes "1" or "0"
 * to /sys/devices/system/cpu/cpu[id]/online.
 *
 * @param thread
 * @param state
 * @returns 0 on success, !0 otherwise
 */
int
set_thread_state(struct thread_info *thread, int state)
{
	char path[DR_PATH_MAX];
	FILE *file;
	int rc = 0;

	sprintf(path, DR_THREAD_ONLINE_PATH, thread->id);
	file = fopen(path, "w");
	if (file == NULL)
		return -1;

	fprintf(file, "%d", state);

	fclose(file);

	/* fprintf apparently does not return negative number
	 * if the write() gets an -EBUSY, so explicitly check the
	 * thread state.
	 */
	if (state != get_thread_state(thread))
		rc = 1;

	return rc;
}

/**
 * get_thread_state
 * @brief Get the "online" status of the given thread.
 *
 * @param thread
 * @returns 0 = offline, 1 = online, -1 on error
 */
int
get_thread_state(struct thread_info *thread)
{
	char path[DR_PATH_MAX];
	int rc, status = -1;

	sprintf(path, DR_THREAD_ONLINE_PATH, thread->id);

	rc = get_int_attribute(path, NULL, &status, sizeof(status));

	return rc ? rc : status;
}

/**
 * cpu_enable_smt
 *
 * For the given physical cpu, online all of its threads.
 * Just call set_cpu_state for now.
 *
 * @param cpu cpu to enable smt on
 * @returns 0 on success, !0 otherwise
 */
int
cpu_enable_smt(struct cpu_info *cpu, struct dr_info *dr_info)
{
	return set_cpu_state(cpu, 1, dr_info);
}

/**
 * cpu_diable_smt
 * @brief Disable all but one of a cpu's threads
 *
 * @param cpu cpu to disable smt on
 * @returns 0 on success, !0 otherwise
 */
int
cpu_disable_smt(struct cpu_info *cpu, struct dr_info *dr_info)
{
	int rc = 0;
	struct thread_info *t;
	int survivor_found = 0;

	for (t = dr_info->all_threads; t != NULL; t = t->next) {
		if (! thread_to_cpu(t, dr_info))
			continue;

		if (1 == get_thread_state(t)) {
			if (survivor_found)
				rc |= set_thread_state(t, 0);
			survivor_found = 1;
		}
	}

	return rc;
}

/**
 * online_first_dead_cpu
 * @brief Find the first cpu with attributes online and physical_id both
 *        set to 0 and online it.
 *
 * @param nthreads
 * @returns 0 on success, !0 otherwise
 */
int
online_first_dead_cpu(int nthreads, struct dr_info *dr_info)
{
	struct thread_info *t;
	int rc = 1;

	for (t = dr_info->all_threads; t != NULL; t = t->next) {
		if (0 == get_thread_state(t) &&
		    0 == thread_physid(t)) {
			    /* just assume that there will
			     * be nthreads to online
			     */
			    while (nthreads--) {
				    dbg("onlining thread %d\n", t->id);
				    set_thread_state(t, 1);
				    t = t->next;
			    }
			    rc = 0;
			    break;
		}
	}

	if (rc)
		err_msg("Could not find any threads to online\n");

	return rc;
}

/**
 * set_cpu_state
 * @brief Given a physical cpu, set all of its threads to the given state.
 *
 * @param cpu
 * @param state
 * @returns 0 on success, !0 otherwise
 */
int
set_cpu_state(struct cpu_info *cpu, int state, struct dr_info *dr_info)
{
	int rc = 0;
	struct thread_info *t = NULL;
	int nr_threads = cpu_nr_threads(cpu);
	int found = 0;

	dbg("setting cpu %s state to %d (%d threads)\n",
	    cpu->name, state, nr_threads);

	/* Hack to work around kernel brain damage (LTC 7692) */
	for (t = dr_info->all_threads; t != NULL; t = t->next) {
		if (thread_to_cpu(t, dr_info) == cpu) {
			found = 1;
			break;
		}
	}

	if (! found) {
		/* There are no threads which match this cpu because
		 * the physical_id attribute is not updated until the
		 * cpu is onlined -- this case is for cpus which are
		 * not present at boot but are added afterwards.
		 */
		return online_first_dead_cpu(nr_threads, dr_info);
	}

	for (t = dr_info->all_threads; t != NULL; t = t->next) {
		if (thread_to_cpu(t, dr_info) != cpu)
			continue;

		if (state != get_thread_state(t))
			rc |= set_thread_state(t, state);

		if (--nr_threads == 0)
			break;
	}

	return rc;
}

/**
 * smt_enabled
 * @brief Is smt enabled?
 *
 * @returns 1 if smt enabled, 0 if not
 */
int
smt_enabled(struct dr_info *dr_info)
{
	/* Just return true if the number of threads in the
	 * first cpu is more than one.
	 */
	if (1 < cpu_nr_threads(dr_info->all_cpus))
		return 1;
	return 0;
}

/**
 * system_enable_smt
 * @brief Activate all threads of each cpu
 *
 * @returns 0 on success, !0 otherwise
 */
int system_enable_smt(struct dr_info *dr_info)
{
	struct cpu_info *cpu;
	int rc = 0;

	for (cpu = dr_info->all_cpus; cpu != NULL; cpu = cpu->next)
		rc |= cpu_enable_smt(cpu, dr_info);

	return rc;
}

/**
 * system_disable_smt
 * @brief Offline all but one thread of each cpu
 *
 * @returns 0 on success, !0 otherwise
 */
int
system_disable_smt(struct dr_info *dr_info)
{
	struct cpu_info *cpu;
	int rc = 0;

	for (cpu = dr_info->all_cpus; cpu != NULL; cpu = cpu->next)
		rc |= cpu_disable_smt(cpu, dr_info);

	return rc;
}
