/**
 * @file drmgr.c
 *
 *
 * Copyright (C) IBM Corporation 2006
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include "dr.h"

#define DRMGR_ARGS	"ac:d:Iimnp:Qq:Rrs:w:"

int debug = 0;
int debug_fd = 0;
char *cmd;

typedef int (cmd_func_t)(struct options *);
typedef int (cmd_args_t)(struct options *);

struct command {
	int		type;
	char		*name;
	cmd_func_t	*func;
	cmd_args_t	*validate_options;
};

#define DRSLOT_CHRP_SLOT	1
#define DRSLOT_CHRP_PHB		2
#define DRSLOT_CHRP_PCI		3
#define DRSLOT_CHRP_MEM		4
#define DRSLOT_CHRP_HEA		5
#define DRSLOT_CHRP_CPU		6
#define DRMIG_CHRP_PMIG		7

static struct command commands[] = {
	{ .type = 0,
	  .name = NULL,
	  .func = NULL,
	  .validate_options = NULL
	},
	{ .type = DRSLOT_CHRP_SLOT,
	  .name = "drslot_chrp_slot",
	  .func = drslot_chrp_slot,
	  .validate_options = valid_slot_options
	},
	{ .type = DRSLOT_CHRP_PHB,
	  .name = "drslot_chrp_phb",
	  .func = drslot_chrp_phb,
	  .validate_options = valid_phb_options
	},
	{ .type = DRSLOT_CHRP_PCI,
	  .name = "drslot_chrp_pci",
	  .func = drslot_chrp_pci,
	  .validate_options = valid_pci_options
	},
	{ .type = DRSLOT_CHRP_MEM,
	  .name = "drslot_chrp_mem",
	  .func = drslot_chrp_mem,
	  .validate_options = valid_mem_options
	},
	{ .type = DRSLOT_CHRP_HEA,
	  .name = "drslot_chrp_hea",
	  .func = drslot_chrp_hea,
	  .validate_options = valid_hea_options
	},
	{ .type = DRSLOT_CHRP_CPU,
	  .name = "drslot_chrp_cpu",
	  .func = drslot_chrp_cpu,
	  .validate_options = valid_cpu_options
	},
	{ .type = DRMIG_CHRP_PMIG,
	  .name = "drmig_chrp_pmig",
	  .func = drmig_chrp_pmig,
	  .validate_options = valid_pmig_options
	},
};

static void
usage()
{
	fprintf(stderr, "Usage: %s {-a | -r} -c connector_type [-s drc_names] "
		"[-f] [-w minutes] [-d detail_level]\n", cmd);
}

int
parse_options(int argc, char *argv[], struct options *opts)
{
	int c;	
	int action_cnt = 0;

	memset(opts, 0, sizeof(*opts));

	/* default timeout */
	opts->timeout = DR_WAIT;

	/* disable getopt error messages */
	opterr = 0;

	while ((c = getopt(argc, argv, DRMGR_ARGS)) != -1) {
		switch (c) {
		    case 'a':
			opts->action = ADD;
			action_cnt++;
			break;
		    case 'c':
			opts->ctype = optarg;
			break;
		    case 'd':
			set_debug(atoi(optarg));
			break;
		    case 'I':
			opts->no_ident = 1;
			break;
		    case 'i':
			opts->action = IDENTIFY;
			action_cnt++;
			break;
		    case 'n':
			opts->noprompt = 1;
			break;
		    case 'p':
			opts->p_option = optarg;
			break;
		    case 'q':
			opts->quantity = strtol(optarg, NULL, 0);
			if (opts->quantity <= 0) {
				err_msg("Invalid quantity %s\n", optarg);
				return -1;
			}
			break;
		    case 'R':
			opts->action = REPLACE;
			action_cnt++;
			break;
		    case 'r':
			opts->action = REMOVE;
			action_cnt++;
			break;
		    case 's':
			opts->usr_drc_name = optarg;
			break;
		    case 'Q':
			opts->action = QUERY;
			action_cnt++;
			break;
		    case 'm':
			opts->action = MIGRATE;
			action_cnt++;
			break;
		    case 'w':
			opts->timeout = strtoul(optarg, NULL, 10) * 60;
			if (opts->timeout < 0) {
				err_msg("Invalid timeout specified: %s\n",
					optarg);
				return -1;
			}
			break;
		    default:
			err_msg("Invalid option specified '%c'\n", c);
			usage();
			return -1;
			break;
		}
	}

	/* Generic parameter checking */
	if (action_cnt == 0) {
		err_msg("At least one action must be specified\n");
		return -1;
	}

	if (action_cnt > 1) {
		err_msg("Only one action may be specified\n");
		return -1;
	}
	
	if ((opts->quantity > 1) && (opts->usr_drc_name)) {
		err_msg("The -q and -s flags are mutually exclusive\n");
		return -1;
	}

	if (opts->ctype == NULL) {
		err_msg("A connector type (-c) must be specified\n");
		return -1;
	}

	return 0;
}

struct command *
get_command(struct options *opts)
{
	/* Unfortunately, the connector type specified doesn't always result
	 * in a 1-to-1 relationship with the resulting command to run so we
	 * have to do some extra checking to build the correct command.
	 */
	if (opts->action == MIGRATE)
		return &commands[DRMIG_CHRP_PMIG];

	if ((! strncmp(opts->ctype, "port", 4)) ||
	    ((opts->usr_drc_name) && (! strncmp(opts->usr_drc_name, "HEA", 3))))
		return &commands[DRSLOT_CHRP_HEA];
	
	if (! strcmp(opts->ctype, "slot"))
		return &commands[DRSLOT_CHRP_SLOT];
			
	if (! strcmp(opts->ctype, "phb"))
		return &commands[DRSLOT_CHRP_PHB];
			
	if (! strcmp(opts->ctype, "pci"))
		return &commands[DRSLOT_CHRP_PCI];
			
	if (! strcmp(opts->ctype, "mem"))
		return &commands[DRSLOT_CHRP_MEM];
			
	if (! strcmp(opts->ctype, "cpu"))
		return &commands[DRSLOT_CHRP_CPU];
			
	/* If we make it this far, the user specified an invalid
	 * connector type.
	 */
	err_msg("Dynamic reconfiguration is not supported for connector\n"
		"type \"%s\" on this system\n", opts->ctype);

	return NULL;
}

int
main(int argc, char *argv[])
{
	struct options opts;
	char log_msg[DR_PATH_MAX];
	struct command *command;
	char *slash;
	int i, offset;
	int lock_id;
	int rc;

	/* We need to know the name of the command that was actually invoked
	 * so grab that out of argv[0]
	 */
	slash = strrchr(argv[0], '/');
	if (slash)
		cmd = slash + 1;
	else
		cmd = argv[0];

	rc = parse_options(argc, argv, &opts);
	if (rc)
		return rc;

	command = get_command(&opts);
	if (command == NULL)
		return -1;

	/* Validate the options for the action we want to perform */
	rc = command->validate_options(&opts);
	if (rc)
		return rc;

	/* Validate this platform */
	if (! valid_platform("chrp"))
		exit(1);

	/* Mask signals so we do not get interrupted */
	if (mask_signals()) {
		err_msg("Could not mask signals to avoid interrupts\n"); 
		return -1;
	}
	
	lock_id = dr_lock(opts.timeout);
	if (lock_id == -1) {
		err_msg("Unable to obtain DR lock. Please try command again "
			"later.\n");
		return -1;
	}

	/* Log this invocation of drmgr */
	offset = sprintf(log_msg, "drmgr: %s ", command->name);
	for (i = 1; i < argc; i++)
		offset += sprintf(log_msg + offset, "%s ", argv[i]);
	log_msg[offset] = '\0';
	syslog(LOG_LOCAL0 | LOG_INFO, log_msg);
	if (debug >= 4) {
		log_msg[offset] = '\n';
		write(debug_fd, log_msg, offset+1);
	}

	/* Now, using the actual command, call out to the proper handler */
	rc = command->func(&opts);

	dr_unlock(lock_id);
	clear_debug();
	return rc;
}

	
