/**
 * @file drmig_chrp_pmig.c
 *
 * Copyright (C) IBM Corporation
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <librtas.h>
#include "dr.h"
#include "drpci.h"


#define OFDT_PMIG_FILE	"/proc/device-tree/ibm,migratable-partition"

struct pmap_struct {
	struct pmap_struct	*next;
	unsigned int		phandle;
	int			ibmphandle;
	char			*name;
};

static struct pmap_struct *plist;

/**
 * add_phandle
 *
 * @param name
 * @param phandle
 * @param ibmphandle
 */
static void
add_phandle(char *name, unsigned int phandle, int ibmphandle)
{
	struct pmap_struct *pm = malloc(sizeof(struct pmap_struct));

	if (strlen(name) == 0)
		name = "/";

	pm->name = malloc(strlen(name)+1);
	strcpy(pm->name,name);

	pm->phandle = phandle;
	pm->ibmphandle = ibmphandle;
	pm->next = plist;
	plist = pm;
}

/**
 * find_phandle
 *
 * @param ph
 * @returns
 */
static char *
find_phandle(unsigned int ph)
{
	struct pmap_struct *pms = plist;

	while (pms && pms->phandle != ph)
		pms = pms->next;

	return pms ? pms->name : NULL;
}

/**
 * add_phandles
 *
 * @param parent
 * @param p
 * @returns
 */
static int
add_phandles(char *parent, char *p)
{
	DIR *d;
	struct dirent *de;
	char path[PATH_MAX];
	unsigned int phandle;
	char *pend;
	FILE *fd;

	strcpy(path,parent);
	if (strlen(p)) {
		strcat(path,"/");
		strcat(path,p);
	}
	pend = path + strlen(path);

	d = opendir(path);
	if (d == NULL) {
		perror(path);
		return 1;
	}

	while ((de = readdir(d))) {
		if ((de->d_type == DT_DIR) &&
		    (strcmp(de->d_name,".")) &&
		    (strcmp(de->d_name,"..")))
			add_phandles(path,de->d_name);
	}

	strcpy(pend,"/linux,phandle");
	fd = fopen(path,"r");
	if (fd != NULL) {
		if (fread(&phandle,sizeof(phandle),1,fd) != 1) {
			perror(path);
			dbg("Error reading phandle data!\n");
			return 1;
		}
		*pend = '\0';
		add_phandle(path + strlen("/proc/device-tree"),phandle, 0);
		fclose(fd);
	}

	strcpy(pend,"/ibm,phandle");
	fd = fopen(path,"r");
	if (fd != NULL) {
		if (fread(&phandle,sizeof(phandle),1,fd) != 1) {
			perror(path);
			dbg("Error reading phandle data!\n");
			return 1;
		}
		*pend = '\0';
		add_phandle(path + strlen("/proc/device-tree"), phandle, 1);
		fclose(fd);
	}

	closedir(d);
	return 0;
}

/**
 * do_update
 *
 * @param cmd
 * @param len
 * @returns 0 on success, !0 otherwise
 */
static int
do_update(char *cmd, int len)
{
	int rc;
	int i, fd;

	fd = open(OFDTPATH, O_WRONLY);
	if (fd <= 0) {
		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
		rc = errno;
		return rc;
	}

	dbg("len %d\n", len);

	if ((rc = write(fd, cmd, len)) != len)
		dbg("Error writing to ofdt file! rc %d errno %d\n", rc, errno);

	for (i = 0; i < len; i++) {
		if (! isprint(cmd[i]))
			cmd[i] = '.';
		if (isspace(cmd[i]))
			cmd[i] = ' ';
	}
	cmd[len] = 0x00;

	dbg("<%s>\n", cmd);

	close(fd);
	return rc;
}

/**
 * del_node
 *
 * @param phandle
 */
static void
del_node(unsigned int phandle)
{
	char *name = find_phandle(phandle);
	char delcmd[128] = "remove_node ";

	if (name == NULL)
		dbg("Delete node error: Invalid phandle %8.8x", phandle);
	else {
		strcat(delcmd,name);
		do_update(delcmd, strlen(delcmd));
	}
}

/**
 * update_properties
 *
 * @param phandle
 * @returns 0 on success, !0 otherwise
 */
static int
update_properties(unsigned int phandle)
{
	int rc;
	char cmd[DR_PATH_MAX];
	char *longcmd = NULL;
	char *newcmd;
	int cmdlen = 0;
	int proplen = 0;
	unsigned int wa[1024];
	unsigned int *op;
	unsigned int nprop;
	unsigned int vd;
	int lenpos = 0;
	char *pname;
	int i;
	int more = 0;
	char *name = find_phandle(phandle);

	memset(wa, 0x00, 16);
	wa[0] = phandle;

	do {
		dbg("about to call rtas_update_properties.  work area:\n");
		dbg("phandle %8.8x, node %s\n",phandle, name ? name : "NULL");
		dbg(" %8.8x %8.8x %8.8x %8.8x\n", wa[0], wa[1], wa[2], wa[3]);

		rc = rtas_update_properties((char *)wa, 1);
		if (rc && rc != 1) {
			dbg("Error %d from rtas_update_properties()\n", rc);
			return 1;
		}

		dbg("successful rtas_update_properties (more %d)\n", rc);

		op = wa+4;
		nprop = *op++;

		for (i = 0; i < nprop; i++) {
			pname = (char *)op;
			op = (unsigned int *)(pname + strlen(pname) + 1);
			vd = *op++;

			switch (vd) {
			    case 0x00000000:
				dbg("%s - name only property %s\n",
				    name, pname);
				break;

			    case 0x80000000:
				dbg("%s - delete property %s\n", name, pname);
				sprintf(cmd,"remove_property %u %s",
					phandle, pname);
				do_update(cmd, strlen(cmd));
				break;

			default:
				if (vd & 0x80000000) {
					dbg("partial property!\n");
					/* twos compliment of length */
					vd = ~vd + 1;
					more = 1;
				}
				else {
					more = 0;
				}

				dbg("%s - updating property %s length %d\n",
				    name, pname, vd);

				/* See if we have a partially completed
				 * command
				 */
				if (longcmd) {
					newcmd = malloc(cmdlen + vd);
					memcpy(newcmd, longcmd, cmdlen);
					free(longcmd);
					longcmd = newcmd;
				}
				else {
					longcmd = malloc(vd+128);
					/* Build the command with a length
					 * of six zeros 
					 */
					lenpos = sprintf(longcmd, 
							 "update_property %u "
							 "%s ", phandle, 
							 pname);
					strcat(longcmd, "000000 ");
					cmdlen = strlen(longcmd);
				}

				memcpy(longcmd + cmdlen, op, vd);
				cmdlen += vd;
				proplen += vd;

				if (! more) {
				        /* Now update the length to its actual
					 * value  and do a hideous fixup of 
					 * the new trailing null 
					 */
					sprintf(longcmd+lenpos,"%06d",proplen);
					longcmd[lenpos+6] = ' ';

					do_update(longcmd, cmdlen);
					free(longcmd);
					longcmd = NULL;
					cmdlen = 0;
					proplen = 0;
				}

				op = (unsigned int *)(((char *)op) + vd);
			}
		}
	} while (rc == 1);

	return 0;
}

/**
 * add_new_node
 *
 * @param phandle
 * @param drcindex
 */
static void
add_new_node(unsigned int phandle, unsigned int drcindex)
{
	char *path;
	int rtas_rc;
	struct of_node *new_nodes;/* nodes returned from configure_connector */

	new_nodes = configure_connector(drcindex);
	
	path = find_phandle(phandle);
	if (path == NULL) {
		dbg("Cannot find pnahdle %x\n", phandle);
		return;
	}

	rtas_rc = add_nodes_to_kernel(path, new_nodes);
	if (rtas_rc)
		dbg("add_nodes_to_kernel failed at %s\n", path);
}

/**
 * del_nodes
 *
 * @param op
 * @param n
 */
static void
del_nodes(unsigned int *op, unsigned int n)
{
	unsigned int phandle;
	int i;

	for (i = 0; i < n; i++) {
		phandle = *op++;
		dbg("Delete node with phandle %8.8x\n", phandle);
		del_node(phandle);
	}
}

/**
 * update_nodes
 *
 * @param op
 * @param n
 */
static void
update_nodes(unsigned int *op, unsigned int n)
{
	unsigned int phandle;
	int i;

	for (i = 0; i < n; i++) {
		phandle = *op++;
		dbg("Update node with phandle %8.8x\n", phandle);
		update_properties(phandle);
	}
}

/**
 * add_nodes
 *
 * @param op
 * @param n
 */
static void
add_nodes(unsigned int *op, unsigned int n)
{
	unsigned int pphandle;
	unsigned int drcindex;
	int i;

	for (i = 0; i < n; i += 2) {
		pphandle = *op++;
		drcindex = *op++;
		dbg("Add node with parent phandle %8.8x and drc index %8.8x\n",
		    pphandle, drcindex);
		add_new_node(pphandle, drcindex);
	}
}

/**
 * devtree_update
 *
 */
static void
devtree_update()
{
	int rc;
	unsigned int wa[1024];
	unsigned int *op;

	if (add_phandles("/proc/device-tree",""))
		return;

	/* First 16 bytes of work area must be initialized to zero */
	memset(wa, 0x00, 16);

	do {
		rc = rtas_update_nodes((char *)wa, 1);
		if (rc && rc != 1) {
			dbg("Error %d from rtas_update_nodes()\n", rc);
			return;
		}

		dbg("successful rtas_update_nodes (more %d)\n", rc);

		op = wa+4;

		while (*op & 0xFF000000) {
			int i;
			dbg("op %p, *op %8.8x\n", op, *op);

			for (i = 0; i < (*op & 0x00FFFFFF); i++)
				dbg("   %8.8x\n",op[i+1]);

			switch (*op & 0xFF000000) {
			    case 0x01000000:
				del_nodes(op+1, *op & 0x00FFFFFF);
				break;

			    case 0x02000000:
				update_nodes(op+1, *op & 0x00FFFFFF);
				break;

			    case 0x03000000:
				add_nodes(op+1, *op & 0x00FFFFFF);
				break;

			    case 0x00000000:
				/* End */
				break;

			    default:
				dbg("Unknown update_nodes op %8.8x\n", *op);
			}
			op += 1 + (*op & 0x00FFFFFF);
		}
	} while (rc == 1);

	dbg("leaving\n");
}

int
valid_pmig_options(struct options *opts)
{
	if (opts->action != MIGRATE) {
		err_msg("The only valid action is -m\n");
		return -1;
	}

	if (opts->p_option  == NULL) {
		err_msg("A command must be specified\n");
		return -1;
	}

	if (opts->usr_drc_name == NULL) {
		err_msg("No streamid specified\n");
		return -1;
	}

	return 0;
}
	
int
drmig_chrp_pmig(struct options *opts)
{
	int rc;
	char *cmd = opts->p_option;
	uint64_t stream_val = strtoull(opts->usr_drc_name, NULL, 16);

	/* Ensure that this partition is migratable/mobile */
	if (kernel_check(OFDT_PMIG_FILE)) {
		fprintf(stderr, "drmig_chrp_pmig: Partition Mobility is not "
			"supported on this kernel.\n");
		return -1;
	}

	/* Today we do no pre-checks for migratability. The only check
	 * we could do is whether the "ibm,suspend-me" RTAS call exists.
	 * But if it doesn't, the firmware level doesn't support migration,
	 * in which case why the heck are we being invoked anyways.
	 */
	dbg("drmig_chrp_pmig: command %s\n",cmd);
	if (strcmp(cmd, "check") == 0)
		return 0;

	/* The only other command is pre */
	if (strcmp(cmd, "pre"))
		return 1;

	/* Now do the actual migration */
	do {
		dbg("drmig_chrp_pmig: about to issue ibm,suspend-me(%llx)\n",
		    stream_val);
		rc = rtas_suspend_me(stream_val);
		dbg("drmig_chrp_pmig: ibm,suspend-me(%llx) = %d\n",
		    stream_val, rc);

		if (rc == NOT_SUSPENDABLE)
			sleep(1);

	} while (rc == NOT_SUSPENDABLE);

	syslog(LOG_LOCAL0 | LOG_INFO, "drmgr: ibm,suspend-me rc %d\n",
	       rc);
	if (rc)
		return rc;

	dbg("drmig_chrp_pmig: Updating device_tree\n");
	devtree_update();

	return 0;
}
