/**
 * @file drpci.h
 *
 *
 * Copyright (C) IBM Corporation
 */

#ifndef _DRPCI_H_
#define _DRPCI_H_

#include "rtas_calls.h"

/*  The following structures are used to keep track of all information
 *  relating to the hot plug PCI slots in the system.  Both Open
 *  Firmware and device information is gathered for each slot.
 *
 * NOTES:
 * 	1) The PCI hot plug slot list is sorted by RPA physical location code.
 *	2) The per-slot Open Firmware node list is sorted by full RPA
 *	   physical location code (slot location code plus connector).
 *	3) The slot is empty if slot.first_node is NULL.
 *	4) A PCI hot plug slot MUST be ignored for PCI hot plug operations if
 *	   slot.skip == 1.  The Open Firmware information is
 *	   not dependable in this case.
 */

/* Error Exit Codes */
#define RC_NONEXISTENT 		3
#define RC_DONT_OWN		4
#define RC_ALREADY_OWN		5
#define RC_LINUX_SLOT		6 /* Special case for ConcMaint */

/* PCI Hot Plug  defs  */
#define PHP_SYSFS_ADAPTER_PATH	"/sys/bus/pci/slots/%s/adapter"
#define PHP_SYSFS_POWER_PATH	"/sys/bus/pci/slots/%s/power"
#define PHP_CONFIG_ADAPTER	1
#define PHP_UNCONFIG_ADAPTER	0

/* The fillowing defines are used for adapter status */
#define EMPTY		0
#define CONFIG		1
#define NOT_CONFIG	2

/* slot type definitions */
#define PHPSLOT	0 /* PCI Hotplug slot */
#define DRSLOT	1 /* logical DR slot */

/* Device type definitions */
#define PCI_HP_DEV	1
#define PCI_DLPAR_DEV	2
#define VIO_DEV		3
#define HEA_DEV		4
#define HEA_PORT_DEV	5
#define PHB_DEV		7

#define DLPAR_ADD_SLOT    	"/sys/bus/pci/slots/control/add_slot"
#define DLPAR_REMOVE_SLOT    	"/sys/bus/pci/slots/control/remove_slot"

#define HEA_ADD_SLOT		"/sys/bus/ibmebus/probe"
#define HEA_REMOVE_SLOT		"/sys/bus/ibmebus/remove"
/* %s is the loc-code of the HEA adapter for *_PORT defines */
#define HEA_ADD_PORT		"/sys/bus/ibmebus/devices/%s/probe_port"
#define HEA_REMOVE_PORT		"/sys/bus/ibmebus/devices/%s/remove_port"

/* This structure represents a PCI device connected to a PCI hot plug slot */
struct pci_info {			/* One per PCI device in HP slot */
	uint            vendor_id;	/* vendor ID */
	uint            device_id;	/* device ID */
	uint            class_code;	/* class code */
};

struct hea_info {
	uint		port_no;
	uint		port_tenure;
};

/* This structure represents a PCI hot plug slot. Data from
 * the Open Firmware tree is gathered here. There is a pointer
 * to a linked list of the OF nodes representing the device(s)
 * connected to this slot (if they exist)
 */
struct node {				/* One per PCI hot plug slot */
	struct node	*next;	/* Next PCI hot plug slot */
	struct node	*children;	/* OF children for this slot */

	int		drc_index;	/* DRC index for slot */
	char		*drc_type;	/* DRC type for slot */
	char		*drc_name;
	int		drc_power;	/* Power domain for slot */

	char		*loc_code;
	char		*ofdt_path;
	char		*ofdt_dname;
	char		*sysfs_dev_path;
	char		*sysfs_hp_path;

	int		skip;
	uint32_t	dev_type;
	int		is_owned;

	union {
		struct pci_info *_pci_info;
		struct hea_info *_hea_info;
		char		*_ic_ofdt_path;
	} _node_u;
#define node_pci_info		_node_u._pci_info
#define node_hea_info		_node_u._hea_info
#define node_ic_ofdt_path	_node_u._ic_ofdt_path
};

#define PCI_NODES	0x00000001
#define VIO_NODES	0x00000002
#define HEA_NODES	0x00000004
#define PHB_NODES	0x00000010

struct node *get_hp_nodes(void);
struct node *get_dlpar_nodes(uint32_t);
struct node *get_node_by_name(const char *, uint32_t, struct node **);
void free_node(struct node *);

/* Function prototypes for subroutines  */
int get_hp_adapter_status(char *);
int set_hp_adapter_status(uint, char *);
int release_hp_children(char *);
int dlpar_io_kernel_op(char *, char *);
int cmp_drcname(char *, char *);
char *get_linux_dname(char *);
char *get_sysfs_hp_path(char *);
int acquire_hp_children(char *, int *);
int enable_hp_children(char *);
int disable_hp_children(char *);

void print_node_list(struct node *);

#endif				/* _DRPCI_H_ */
