/**
 * @file drslot_chrp_hea.c
 *
 * Copyright (C) IBM Corporation 2006
 *
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <librtas.h>
#include <dirent.h>

#include "dr.h"
#include "drpci.h"
#include "ofdt.h"

#define PORT_ADD	0
#define PORT_REMOVE	1

static void
usage()
{
	fprintf(stderr, "Usage: %s {-a | -r | -Q} -c slot|port -s drc_name\n",
		cmd);
}

/**
 * sysfs_hea_op
 * @brief Write the proper information to the specified sysfs file for HEA
 *
 * @param fname sysfs file to write to
 * @param name information to write to the sysfs file
 * @returns 0 on success, !0 otherwise
 */
static int
sysfs_hea_op(char *fname, char *name)
{
	int rc = 0, len;
	FILE *file;

    	file = fopen(fname, "w");
    	if (file == NULL) {
		err_msg("%s %s\n", fname, strerror(errno));
		return -ENODEV;
    	}

	len = strlen(name);
    	rc = fwrite(name, 1, len, file);
	fclose(file);

	return (rc >= 0) ? 0 : rc;
}

/**
 * sysfs_port_op
 * @brief Write the port number to sysfs for HEA port add/remove
 *
 * @param action add/remove of the port
 * @parm slot slot to add/remove
 * @return 0 on success, !0 otherwise
 */
static int
sysfs_port_op(int action, struct node *hea, struct node *port)
{
	char fname[128];
	char port_str[4];
	uint port_no = -1;

	if (! hea->sysfs_dev_path)
		return -EINVAL;

	if (action == 0)
		sprintf(fname, "%s/%s", hea->sysfs_dev_path, "probe_port");
	else
		sprintf(fname, "%s/%s", hea->sysfs_dev_path, "remove_port");

	get_property(port->ofdt_path, "ibm,hea-port-no", &port_no,
		     sizeof(port_no));
	if (port_no == -1)
		return -EINVAL;

	sprintf(port_str, "%d", port_no);

	return sysfs_hea_op(fname, port_str);
}

/**
 * remove_port
 * @brief Remove the HEA adapater or port specified by the options
 *
 * @param opts command options
 * @returns 0 on success, !0 otherwise
 */
static int
remove_port(struct options *opts)
{
	struct node *hea, *all_nodes;
	struct node *port, *tmp;
	int rc;
	int no_ports = 0;
	char cmd[256];

	hea = get_node_by_name(opts->usr_drc_name, HEA_NODES, &all_nodes);
	if (hea == NULL)
		return RC_NONEXISTENT;

	for (port = hea->children; port; port = port->next) {
		if (! strcmp(port->drc_name, opts->usr_drc_name))
			break;
	}

	dbg("The sensor-state of \"%s\" is %d\n", port->drc_name,
	     dr_entity_sense(port->drc_index));

	/* need to remove slot from sysfs which will "Hot unplug" the
	 * slot from pci world. unmap_bus_range will be done here also.
	 */
	dbg("Removing \"%s\" from hotplug subsystem\n", port->drc_name);
	rc = sysfs_port_op(PORT_REMOVE, hea, port);
	if (rc) {
		dbg("remove \"%s\" from hotplug subsystem failed\n",
		    port->drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		free_node(all_nodes);
		return -1;
	}

	/* if this is the last port to be removed, we need to remove the
	 * hea adapter from the OS also.  do not actually remove the
	 * HEA adapter from the device tree or unallocate it, we will get
	 * a request to do that later.
	 */
	for (tmp = hea->children; tmp; tmp = tmp->next)
		no_ports++;

	if (no_ports == 1) {
		dbg("Removing \"%s\" from hotplug subsystem\n", hea->drc_name);
		rc = sysfs_hea_op(HEA_REMOVE_SLOT,
				  strstr(hea->ofdt_path, "/lhea"));
		if (rc) {
			dbg("remove \"%s\" from hotplug subsystem failed\n",
			    hea->drc_name);
			err_msg("Unknown failure. Data may be out of sync and\n"
				"the system may require a reboot.\n");
			free_node(all_nodes);
			return -1;
		}
	}

	rc = release_drc(port->drc_index, port->dev_type);
	if (rc)
		return rc;

	sprintf(cmd, "/%s", strstr(port->ofdt_path, "lhea"));
	rc = remove_nodes_from_kernel(cmd);
	if (rc) {
		dbg("removing node \"%s\" failed\n", port->drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		free_node(all_nodes);
		return -1;
	}

	dbg("device node(s) for %s removed\n", port->drc_name);
	free_node(all_nodes);
	return 0;
}

/**
 * remove_hea
 * @brief Remove the HEA adapater or port specified by the options
 *
 * @param opts command options
 * @returns 0 on success, !0 otherwise
 */
static int
remove_hea(struct options *opts)
{
	struct node *hea, *all_nodes;
	int rc;
	char cmd[256];

	hea = get_node_by_name(opts->usr_drc_name, HEA_NODES, &all_nodes);
	if (hea == NULL)
		return RC_NONEXISTENT;

	dbg("The sensor-state of \"%s\" is %d\n", hea->drc_name,
	     dr_entity_sense(hea->drc_index));

	rc = release_drc(hea->drc_index, hea->dev_type);
	if (rc) {
		free_node(all_nodes);
		return rc;
	}

	sprintf(cmd, "/%s", strstr(hea->ofdt_path, "lhea"));
	rc = remove_nodes_from_kernel(cmd);
	if (rc) {
		dbg("removing node \"%s\" failed\n", hea->drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		free_node(all_nodes);
		return -1;
	}

	dbg("device node(s) for %s removed\n", hea->drc_name);
	free_node(all_nodes);
	return 0;
}

/**
 * add_slot
 * @bried Add the HEA adapter or port specified by the options
 *
 * @param opts command options
 * @returns 0 on success, !0 otherwise
 */
static int
add_slot(struct options *opts)
{
	struct dr_connector drc;
	struct of_node *new_node;
	struct node *hea;
	struct node *port;
	struct node *all_nodes;
	char ofdt_path[DR_PATH_MAX];
	char *root_dir;
	int rc = 0;

	rc = get_drc_by_name(opts->usr_drc_name, &drc, ofdt_path, OFDT_BASE);
	if (rc)
		return rc;

	rc = acquire_drc(drc.index, &new_node);
	if (rc)
		return rc;

	root_dir = strstr(ofdt_path, "/lhea");
	if (root_dir == NULL)
		root_dir = "/";

	rc = add_nodes_to_kernel(root_dir, new_node);
	if (rc) {
		dbg("add_nodes_to_kernel failed\n");
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");

		drc_recovery(drc.index);
		return -1;
	}

	hea = get_node_by_name(opts->usr_drc_name, HEA_NODES, &all_nodes);
	if (hea == NULL) {
		dbg("Could not get find \"%s\"\n", opts->usr_drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");

		drc_recovery(drc.index);
		return -1;
	}

	switch (opts->usr_drc_name[0]) {
	    case 'H':
		dbg("Adding \"%s\" to hotplug subsystem\n", hea->drc_name);
		rc = sysfs_hea_op(HEA_ADD_SLOT,
				  strstr(hea->ofdt_path, "/lhea"));
		break;
	    case 'P':
		for (port = hea->children; port; port = port->next) {
			if (! strcmp(opts->usr_drc_name, port->drc_name))
				break;
		}

		dbg("Adding \"%s\" to hotplug subsystem\n", port->drc_name);
		rc = sysfs_port_op(PORT_ADD, hea, port);
		break;
	}
	if (rc) {
		dbg("add failed\n");
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");

		drc_recovery(drc.index);
		free_node(all_nodes);
		return -1;
	}

	free_node(all_nodes);
	return 0;
}

int
valid_hea_options(struct options *opts)
{
	if (opts->usr_drc_name == NULL) {
		err_msg("A drc name  must be specified\n");
		return -1;
	}

	return 0;
}

int
drslot_chrp_hea(struct options *opts)
{
	int rc;

	if (kernel_check(HEA_ADD_SLOT)) {
		err_msg("DLPAR HEA operations are not supported on"
			"this kernel");
		return -1;
	}

	switch (opts->action) {
	    case ADD:
		rc = add_slot(opts);
		break;

	    case REMOVE:
		if (! strcmp(opts->ctype, "port"))
			rc = remove_port(opts);
		else if (! strcmp(opts->ctype, "slot"))
			rc = remove_hea(opts);
		else {
			err_msg("The connector type %s is not supported.\n",
				opts->ctype);
			rc = -1;
		}
		break;

	    case QUERY:
		{
			struct node *node, *all_nodes;
			node = get_node_by_name(opts->usr_drc_name, HEA_NODES,
						&all_nodes);
			if (node == NULL) {
				err_msg("%s not owned by partition\n",
					opts->usr_drc_name);
				rc = RC_DONT_OWN;
			} else {
				/* Special case for HMC */
				rc = RC_LINUX_SLOT;
			}

			free_node(all_nodes);
			break;
		}

	    default:
		usage();
		rc = -1;
	}

	return rc;
}
