/**
 * @file drslot_chrp_mem
 *
 *
 * Copyright (C) IBM Corporation 2006
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/wait.h>
#include "dr.h"
#include "ofdt.h"

#define DR_STR_MAX	128

struct lmb {
	struct lmb	*next;
	uint32_t	flags;

	int		drc_index;
	int		drc_power;
	char		drc_type[DR_STR_MAX];
	char		drc_name[DR_STR_MAX];

	char		ofdt_path[DR_PATH_MAX];
};

#define LMB_OWNED	0x00000001
#define LMB_ACQUIRED	0x00000002
#define LMB_RELEASED	0x00000004
#define LMB_STATE_SET	0x00000008

#define MEM_PROBE_FILE		"/sys/devices/system/memory/probe"
#define MEM_BLOCK_SIZE_BYTES	"/sys/devices/system/memory/block_size_bytes"

static int lmbs_modified = 0;

/**
 * free_lmb_list
 * @brief Free the list of lmb structs
 *
 * @param lmb_list list of structs to free
 */
static void
free_lmb_list(struct lmb *lmb_list)
{
	struct lmb *lmb;

	while (lmb_list != NULL) {
		lmb = lmb_list;
		lmb_list = lmb->next;
		free(lmb);
	}
}

/**
 * get_lmbs
 * @brief Build a list of all possible lmbs for the system
 *
 * @return list of lmbs, NULL on failure
 */
static struct lmb *
get_lmbs()
{
	struct dr_connector *drc_list, *drc;
	struct lmb *lmb_list = NULL;
	struct lmb *lmb;
	struct dirent *de;
	DIR *d;
	int rc = 0;

	drc_list = init_drc_info(OFDT_BASE);
	if (drc_list == NULL)
		return NULL;

	/* For memory dlpar, we need a list of all posiible memory nodes
	 * for the system, initalize those here.
	 */
	for (drc = drc_list; drc; drc = drc->next) {
		if (strncmp(drc->name, "LMB", 3))
			continue;

		lmb = malloc(sizeof(*lmb));
		if (lmb == NULL) {
			free_lmb_list(lmb_list);
			return NULL;
		}

		memset(lmb, 0, sizeof(*lmb));

		lmb->drc_index = drc->index;
		lmb->drc_power = drc->powerdomain;

		strcpy(lmb->drc_type, drc->type);
		strcpy(lmb->drc_name, drc->name);

		lmb->next = lmb_list;
		lmb_list = lmb;
	}

	/* We also need to know which lmbsw are already allocated to
	 * the system.  Walk the device tree and update the appropriate
	 * lmb entries as we find their device tree entries.
	 */
	d = opendir(OFDT_BASE);
	if (d == NULL) {
		free_lmb_list(lmb_list);
		return NULL;
	}

	while ((de = readdir(d)) != NULL) {
		if (de->d_type != DT_DIR)
			continue;

		if (! strncmp(de->d_name, "memory@", 7)) {
			char path[1024];
			int my_drc_index;

			memset(path, 0, 1024);
			sprintf(path, "%s/%s", OFDT_BASE, de->d_name);

			if (get_my_drc_index(path, &my_drc_index))
				continue;

			for (lmb = lmb_list; lmb; lmb = lmb->next) {
				if (lmb->drc_index == my_drc_index)
					break;
			}

			if (lmb == NULL) {
				rc = -1;
				break;
			}

			strcpy(lmb->ofdt_path, path);
			lmb->flags &= LMB_OWNED;
		}
	}

	closedir(d);

	if (rc) {
		free_lmb_list(lmb_list);
		lmb_list = NULL;
	}

	return lmb_list;
}

/**
 * get_lmb_to_add
 *
 * Find the first lmb which does not correspond to a lmb
 * already owned by the partition and is avaialble, or the lmb
 * matching the one specified by the user.
 *
 * @param lmb_list list of lmbs to be searched for available lmb
 * @returns pointer to avaiable lmb on success, NULL otherwise
 */
static struct lmb *
get_lmb_to_add(struct options *opts, struct lmb *lmb_list)
{
	struct lmb *lmb;

	for (lmb = lmb_list; lmb; lmb = lmb->next) {
		int rc;

		if (opts->usr_drc_name) {
			if (lmb->drc_name == opts->usr_drc_name)
				break;
			else
				continue;
		}

		if (lmb->flags & LMB_OWNED)
			continue;

		rc = dr_entity_sense(lmb->drc_index);
		if (rc != STATE_UNUSABLE)
			continue;

		/* Found an available lmb */
		break;
	}

	if (lmb)
		dbg("Found avaialable lmb, drc index 0x%x\n", lmb->drc_index);
	else
		dbg("Could not find available lmb to add\n");

	return lmb;
}

static struct lmb *
get_lmb_to_remove(struct options *opts, struct lmb *lmb_list)
{
	struct lmb *lmb;

	if (! opts->usr_drc_name) {
		err_msg("Memory remove only supported when specifying the drc "
			"name of\nthe lmb to remove.\n");
		return NULL;
	}

	for (lmb = lmb_list; lmb; lmb = lmb->next) {
		if (opts->usr_drc_name) {
			if (lmb->drc_name == opts->usr_drc_name)
				break;
			else
				continue;
		}
	}

	if (lmb)
		dbg("Found avaialable lmb, drc index 0x%x\n", lmb->drc_index);
	else
		dbg("Could not find available lmb to remove\n");

	return lmb;
}

/**
 * acquire_lmb
 * @brief Acquire a new lmb for this partition from the hypervisor.
 *
 * @param lmb lmb to acquire
 * @returns 0 on success, !0 otherwise
 */
static int
acquire_lmb(struct lmb *lmb)
{
        struct of_node *new_node;
	char add_path[DR_PATH_MAX];
        int rc;

	rc = acquire_drc(lmb->drc_index, &new_node);
	if (rc)
		return rc;

        /* Add the new nodes to the device tree */
	sprintf(add_path, "/%s", new_node->name);
	rc = add_node(add_path, new_node);
	if (rc)
		return rc;

	/* update the lmb with ofdt path and ownership */
	sprintf(lmb->ofdt_path, "%s/%s", OFDT_BASE, add_path);
	lmb->flags |= LMB_ACQUIRED;

        return 0;
}

/**
 * release_lmb
 * @brief Remove a lmb from the device tree.
 *
 * @param lmb lmb to release
 * @returns 0 on success, !0 otherwise
 */
static int
release_lmb(struct lmb *lmb)
{
	int rc;

	rc = release_drc(lmb->drc_index, 0);
	if (rc)
		return rc;

        rc = remove_node(lmb->ofdt_path);

	memset(lmb->ofdt_path, 0, DR_PATH_MAX);
	lmb->flags |= LMB_RELEASED;

	return rc;
}

/**
 * set_lmb_state
 *
 * @param lmb lmb to set the state for
 * @param state 1 = online, 0 = offline
 * @returns 0 on success, !0 otherwise
 */
static int
set_lmb_state(struct lmb *lmb, int state_val)
{
	int probe_file;
	uint32_t regs[4];
	uint32_t lmb_sz;
	uint64_t phys_addr;
	uint32_t scn_sz;
	int mem_scn;
	char *tmp;
	char buf[DR_STR_MAX];
	char *state = (state_val ? "online" : "offline");
	int rc;

	rc = get_property(lmb->ofdt_path, "reg", &regs, sizeof(regs));
	if (rc) {
		dbg("Could not determine lmb size for %s\n", lmb->ofdt_path);
		return rc;
	}
	lmb_sz = regs[3];
	
	rc = get_str_attribute("/sys/devices/system/memory", 
			       "/block_size_bytes", &buf, DR_STR_MAX);
	if (rc) {
		dbg("Could not determine section size for %s\n",
		    lmb->ofdt_path);
		return rc;
	}

	scn_sz = strtoul(buf, NULL, 16);

	tmp = strrchr(lmb->ofdt_path, '@');
	if (tmp == NULL) {
		dbg("Could not determine physical address for %s\n",
		    lmb->ofdt_path);
		return rc;
	}

	tmp++;
	phys_addr = strtoull(tmp, NULL, 16);

	mem_scn = phys_addr / scn_sz;
	probe_file = open(MEM_PROBE_FILE, O_WRONLY);
	if (probe_file <= 0) {
		dbg("Could not open %s to probe memory\n", MEM_PROBE_FILE);
		return errno;
	}

	while (lmb_sz > 0) {
		int file;
		char path[DR_PATH_MAX];
		char addr[DR_STR_MAX];
		char *state_str = "/sys/devices/system/memory/memory%d/state";

		memset(addr, 0, DR_STR_MAX);
		sprintf(addr, "0x%llx", phys_addr);

		rc = write(probe_file, addr, strlen(addr));
		if (rc == -1) {
			dbg("Could not write to %s:\n%s\n", MEM_PROBE_FILE,
			    strerror(errno));
			break;
		}

		memset(path, 0, DR_PATH_MAX);
		sprintf(path, state_str, mem_scn);
			
		file = open(path, O_WRONLY);
		if (file <= 0) {
			dbg("Could not open %s to online its memory.\n%s\n",
			    path, strerror(errno));
			rc = errno;
			break;
		}

		rc = write(file, state, strlen(state));
		close(file);
		if (rc == -1) {
			dbg("Could not write to %s:\n%s\n", MEM_PROBE_FILE,
			    path);
			break;
		}

		lmb_sz -= scn_sz;
		phys_addr += scn_sz;
		mem_scn = phys_addr / scn_sz;

		/* reset rc so we don't exit the loop with rc set 
		 * to the return value from the last write() call
		 */
		rc = 0;
	}

	close(probe_file);
	lmb->flags |= LMB_STATE_SET;
	return rc;
}

/**
 * add_lmbs
 *
 * Attempt to acquire and online the given number of LMBs.
 * This function calls itself recursively to simplify recovery
 * actions in case of an error.  This is intended only for the case
 * where the user does not specify a drc-name.
 *
 * @param nr_lmbs number of lmbs to add
 * @param lmb_list list of lmbs on the partition
 * @returns 0 on success, !0 otherwise
 */
static int
add_lmbs(int nr_lmbs, struct options *opts, struct lmb *lmb_list)
{
	int rc = 0;
	struct lmb *lmb;

	lmb = get_lmb_to_add(opts, lmb_list);
	if (lmb == NULL)
                return -1;

	rc = acquire_lmb(lmb);
	if (rc)
		return -1;

	if (nr_lmbs > 1)
		rc = add_lmbs(nr_lmbs - 1, opts, lmb_list);

	if (rc) {
		/* Roll back the operation.  Is this the correct behavior? */
		release_lmb(lmb);
		return 1;
	}

	rc = set_lmb_state(lmb, 1);
	if (rc) {
		/* Roll back the operation.  Is this the correct behavior? */
#if 0
		set_lmb_state(lmb, 0);
#endif
		release_lmb(lmb);
		return 1;
	}

	lmbs_modified++;
	return 0;
}

/**
 * mem_add
 * @brief Add memory to the partition
 *
 * @param opts user options
 * @returns 0 on success, !0 otherwise
 */
static int
mem_add(struct options *opts)
{
	struct lmb *lmb_list;
	int rc;

	lmb_list = get_lmbs();
	if (lmb_list == NULL)
		return -1;

	dbg("Adding %d lmbs\n", opts->quantity);
	rc = add_lmbs(opts->quantity, opts, lmb_list);
	printf("DR_TOTAL_RESOURCES=%d\n", lmbs_modified);

	free_lmb_list(lmb_list);
	return rc;
}

/**
 * remove_lmbs
 *
 * @param nr_lmbs
 * @param opts
 * @param lmb_list
 * @return 0 on success, !0 otherwise
 */
static int
remove_lmbs(int nr_lmbs, struct options *opts, struct lmb *lmb_list)
{
	struct lmb *lmb;
	int rc;

	lmb = get_lmb_to_remove(opts, lmb_list);
	if (lmb == NULL)
		return -1;
	
	rc = set_lmb_state(lmb, 0);
	if (rc)
		return rc;

	if (nr_lmbs > 1)
		rc = remove_lmbs(nr_lmbs - 1, opts, lmb_list);

	if (rc) {
		rc = set_lmb_state(lmb, 1);
		return -1;
	}

	rc = release_lmb(lmb);
	if (rc) {
		acquire_lmb(lmb);
		set_lmb_state(lmb, 1);
		return -1;
	}

	lmbs_modified++;
	return 0;
}

/**
 * mem_remove
 *
 * @param opts
 * @return 0 on success, !0 otherwise
 */
static int
mem_remove(struct options *opts)
{
	struct lmb *lmb_list;
	int rc;

	lmb_list = get_lmbs();
	if (lmb_list == NULL)
		return -1;

	dbg("Removing %d lmbs\n", opts->quantity);
	rc = remove_lmbs(opts->quantity, opts, lmb_list);
	printf("DR_TOTAL_RESOURCES=%d\n", lmbs_modified);

	free_lmb_list(lmb_list);
	return rc;
}

/*********************************************************************
 * Warning: Ugly hacking below that should be removed as early as possible
 *
 * In order to properly support memory DLPAR we have to ensure that
 * the ehea module is either not loaded or we are using a version that
 * can handle memory dlpar.  Otherwise bad stuff happens.
 */

/* These two defines are taken from drivers/net/ehea/ehea.h in the kernel.
 * Unfortunately they do not appear in any header we can include so we
 * have to define the values here so we can check ehea capabilities.
 */
#define MEM_ADD_ATTR	0x0000000000000002
#define MEM_RM_ATTR	0x0000000000000004

/**
 * ehea_compatable
 * @brief Determine if ehea is loaded and if it can handle memory dlpar
 *
 * This uses system() to run lsmod and grep to check for the presence
 * of the ehea module.  If it is present we check its capabilities file
 * to determine if it can handle memory dlpar.
 *
 * @param action memory add/remove action that is to be checked for
 * @return 1 if we are ehea compatable, 0 otherwise
 */
static int
ehea_compatable(int action)
{
	FILE *fp;
	uint64_t flags = 0;
	int rc;

	rc = system("/sbin/lsmod | grep ehea >/dev/null 2>&1");
	if (WEXITSTATUS(rc) != 0) {
		/* The ehea module is not loaded, everything is good */
		return 1;
	}

	/* The module is loaded, now we need to see if it
	 * can handle memory dlpar operations.
	 */
	fp = fopen("/sys/bus/ibmebus/drivers/ehea/capabilities", "r");
	if (fp == NULL) {
		/* File doesn't exist, _not_ memory dlpar compatable */
		return 0;
	}

	rc = fscanf(fp, "%llu", &flags);
	fclose(fp);

	if ((action == ADD) && (flags & MEM_ADD_ATTR))
		return 1;

	if ((action == REMOVE) && (flags & MEM_RM_ATTR))
		return 1;

	return 0;
}

int
valid_mem_options(struct options *opts)
{
	/* default to a quantity of 1 */
	if ((opts->quantity == 0) && (opts->usr_drc_name == NULL))
		opts->quantity = 1;

	return 0;
}

int
drslot_chrp_mem(struct options *opts)
{
	int rc;

	if (kernel_check(MEM_BLOCK_SIZE_BYTES)) {
		err_msg("DLPAR memory operations are not supported on"
			"this kernel.");
		return -1;
	}

	if (! ehea_compatable(opts->action)) {
		err_msg("Memory DLPAR is not supported on partitions "
			"using HEA adapters\n");
		return -1;
	}

	/* The recursive nature of the routines that add/remove lmbs
	 * require that the quantity be non-zero.
	 */
	if (opts->usr_drc_name)
		opts->quantity = 1;

	switch (opts->action) {
	    case ADD:
		rc = mem_add(opts);
		break;

	    case REMOVE:
#ifdef MEM_REMOVE
		rc = mem_remove(opts);
#else
		/* 8-25-04: Joel said that lmb removal isn't working */
		err_msg("The memory removal is not supported!");
		rc = -1;
#endif
		break;
	}

	return rc;
}
