/**
 * @file drslot_chrp_phb.c
 *
 *
 * Copyright (C) IBM Corporation 2006
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <librtas.h>
#include <errno.h>
#include "dr.h"
#include "drpci.h"
#include "ofdt.h"

/**
 * usage_error
 * @brief print the usage message
 *
 * @returns 1, always
 */
static int
usage_error(void)
{
	fprintf(stderr, "Usage: %s [-Q | -r | -a] -c phb -s <phb_name> "
		"[-d level]\n", cmd);
	return 1;
}

static struct node *
find_phb_by_name(struct node *phb_list, const char *drc_name)
{
	struct node *phb;

	for (phb = phb_list; phb; phb = phb->next) {
		if (! strcmp(phb->drc_name, drc_name))
			return phb;
	}

	return NULL;
}

/**
 * query_phb
 *
 * @param op
 * @returns 0 if a remove would succeed, or if it's alreday been removed
 * @returns 1 if a remove would not succeed
 */
static int
query_phb(struct options *opts)
{
	struct node *phb_list;
	struct node *phb;
	struct node *child;

	phb_list = get_dlpar_nodes(PHB_NODES);
	if (phb_list == NULL)
		return -1;

	phb = find_phb_by_name(phb_list, opts->usr_drc_name);
	if (phb == NULL) {
		free_node(phb_list);
		return -1;
	}

	/* If this PHB still owns children that are not hotplug, fail. */
	for (child = phb->children; child; child = child->next) {
		if ((child->is_owned) && (child->dev_type != PCI_HP_DEV)) {
			free_node(phb_list);
			return -1;
		}
	}

	free_node(phb_list);
	return 0;
}

/**
 * remove_phb_of_nodes
 *
 * @param phb
 * @returns 0 opn success, !0 otherwise
 */
static int
remove_phb_of_nodes(struct node *phb)
{
	char path[DR_PATH_MAX];
	int rc;

	sprintf(path, "%s", strstr(phb->ofdt_path, "/pci"));
	rc = remove_nodes_from_kernel(path);
	if (rc)
		return rc;

	sprintf(path, "%s", strstr(phb->node_ic_ofdt_path, "/interrupt"));
	rc = remove_nodes_from_kernel(path);
	if (rc)
		return rc;

	return 0;
}

/**
 * remove_phb
 *
 * @param op
 * @returns 0 on success, !0 otherwise
 */
static int
remove_phb(struct options *opts)
{
	struct node *phb_list = NULL;
	struct node *phb;
	struct node *child;
	int rc = 0;

	phb_list = get_dlpar_nodes(PHB_NODES);
	if (phb_list == NULL)
		return -1;

	phb = find_phb_by_name(phb_list, opts->usr_drc_name);
	if (phb == NULL) {
		rc = -1;
		goto phb_remove_error;
	}

	/* If this PHB still owns children that are not hotplug, fail. */
	for (child = phb->children; child; child = child->next) {
		if ((child->is_owned) && (child->dev_type != PCI_HP_DEV)) {
			rc = -1;
			goto phb_remove_error;
		}
	}

	/* Now, disable any hotplug children */
	for (child = phb->children; child; child = child->next) {
		if (child->dev_type == PCI_HP_DEV) {
			rc = disable_hp_children(child->drc_name);
			if (rc)
				err_msg("failed to disable hotplug children\n");

			rc = release_hp_children(child->drc_name);
			if (rc && rc != -EINVAL) {
				err_msg("failed to release hotplug children\n");
				goto phb_remove_error;
			}
		}
	}

	rc = dlpar_io_kernel_op(DLPAR_REMOVE_SLOT, phb->drc_name);
	if (rc) {
		err_msg("kernel remove failed for %s, rc = %d\n",
			phb->drc_name, rc);
		goto phb_remove_error;
	}

	rc = release_drc(phb->drc_index, phb->dev_type);
	if (rc)
		goto phb_remove_error;

	rc = remove_phb_of_nodes(phb);
	if (rc) {
		err_msg("failed to remove PHB OF nodes\n");
		/* TODO: unroll */
		return 1;
	}

phb_remove_error:
	if (phb_list)
		free_node(phb_list);

	return rc;
}

/**
 * add_phb
 *
 * @param op
 * @returns 0 on success, !0 otherwise
 */
static int
add_phb(struct options *opts)
{
	struct dr_connector drc;
	char root_dir[DR_PATH_MAX];
	struct node *phb_list = NULL;
	struct node *phb;
	struct of_node *new_nodes;
	int rc, n_children = 0;

	rc = get_drc_by_name(opts->usr_drc_name, &drc, root_dir, OFDT_BASE);
	if (rc) {
		err_msg("Could not find drc name \"%s\" to remove PHB\n",
			opts->usr_drc_name);
		goto phb_add_error;
	}

	rc = acquire_drc(drc.index, &new_nodes);
	if (rc)
		goto phb_add_error;

	rc = add_nodes_to_kernel("/", new_nodes);
	if (rc) {
		err_msg("unable to add new device nodes, rc = %d\n", rc);
		goto phb_add_error;
	}

	phb_list = get_dlpar_nodes(PHB_NODES);
	if (phb_list == NULL) {
		dbg("Could not generate PHB list\n");
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		rc = -1;
		goto phb_add_error;
	}

	phb = find_phb_by_name(phb_list, opts->usr_drc_name);
	if (phb == NULL) {
		dbg("Could not get find \"%s\"\n", opts->usr_drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		rc = -1;
		goto phb_add_error;
	}

	rc = acquire_hp_children(strstr(phb->ofdt_path, "/pci"), &n_children);
	if (rc) {
		err_msg("unable to acquire PHB hotplug children\n");
		goto phb_add_error;
	}

	rc = dlpar_io_kernel_op(DLPAR_ADD_SLOT, phb->drc_name);
	if (rc) {
		err_msg("add %s into hotplug subsystem failed\n",
			phb->drc_name);
		goto phb_add_error;
	}

	if (n_children) {
		rc = enable_hp_children(phb->drc_name);
		if (rc)
			err_msg("enable children failed for %s\n",
				phb->drc_name);
	}

phb_add_error:
	if (phb_list)
		free_node(phb_list);

	return rc;
}

int
valid_phb_options(struct options *opts)
{
	if (opts->usr_drc_name == NULL) {
		err_msg("A drc name must be specified\n");
		return -1;
	}

	return 0;
}

int
drslot_chrp_phb(struct options *opts)
{
	int rc;

	if (kernel_check(DLPAR_ADD_SLOT)) {
		err_msg("DLPAR PHB operations are not supported on"
			"this kernel.");
		return -1;
	}

	switch(opts->action) {
	    case ADD:
		rc = add_phb(opts);
		break;
	    case REMOVE:
		rc = remove_phb(opts);
		break;
	    case QUERY:
		rc = query_phb(opts);
		break;
	}

	return rc;
}
