/**
 * @file drslot_chrp_slot.c
 *
 * Copyright (C) IBM Corporation 2006
 *
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <librtas.h>

#include "dr.h"
#include "drpci.h"
#include "ofdt.h"

static void
usage()
{
	fprintf(stderr, "Usage: %s {-a | -r | -Q} -c slot -s drc_name\n", cmd);
}

static int
query_slot(struct options *opts)
{
	struct node *all_nodes;
	struct node *node;
	int rc;

	node = get_node_by_name(opts->usr_drc_name, PCI_NODES | VIO_NODES,
				&all_nodes);
	if (node == NULL)
		return RC_NONEXISTENT;

	if (! node->is_owned) {
		err_msg("%s not owned by partition\n", opts->usr_drc_name);
		rc = RC_DONT_OWN;
	} else
		/* Special case for HMC */
		rc = RC_LINUX_SLOT;

	free_node(all_nodes);
	return rc;
}

/**
 * remove_slot
 *
 */
static int
remove_slot(struct options *opts)
{
	struct node *all_nodes;
	struct node *node;
	int rc;
	char cmd[256], *remove_path;

	node = get_node_by_name(opts->usr_drc_name, PCI_NODES | VIO_NODES,
				&all_nodes);
	if (node == NULL) {
		err_msg("%s does not exist\n", opts->usr_drc_name);
		return RC_NONEXISTENT;
	}

	if (! node->is_owned) {
		err_msg("%s not owned by partition\n", opts->usr_drc_name);
		rc = RC_DONT_OWN;
		goto slot_remove_exit;
	}

	if (node->sysfs_hp_path) {
		rc = disable_hp_children(node->drc_name);
		if (rc)
			err_msg("failed to disable hotplug children\n");

		rc = release_hp_children(node->drc_name);
		if (rc && rc != -EINVAL) {
			err_msg("failed to release hotplug children\n");
			goto slot_remove_exit;
		}
	}

	dbg("The sensor-state of drc_index 0x%x is %d\n",
	     node->drc_index, dr_entity_sense(node->drc_index));

	/* need to remove slot from sysfs which will
	 * "Hot unplug" the slot from pci world. unmap_bus_range will be
	 * done here also.
	 */
	rc = dlpar_io_kernel_op(DLPAR_REMOVE_SLOT, node->drc_name);
	if (rc) {
		dbg("remove %s from hotplug subsystem failed\n",
		    node->drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		goto slot_remove_exit;
	}

	rc = release_drc(node->drc_index, node->dev_type);
	if (rc)
		goto slot_remove_exit;

	switch (node->dev_type) {
	    case PCI_DLPAR_DEV:
		remove_path = strstr(node->ofdt_path, "pci");
		break;

	    case VIO_DEV:
		remove_path = strstr(node->ofdt_path, "vdevice");
		break;
	}

	sprintf(cmd, "/%s", remove_path);
	rc = remove_nodes_from_kernel(cmd);
	if (rc) {
		dbg("removing node %s failed\n", node->drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		goto slot_remove_exit;
	}

	dbg("device node(s) for %s removed\n", node->drc_name);

slot_remove_exit:
	if (all_nodes)
		free_node(all_nodes);

	return rc;
}

/**
 * add_slot
 *
 */
static int
add_slot(struct options *opts)
{
	struct dr_connector drc;
	struct node *all_nodes;
	struct node *node;
	struct of_node *new_nodes;
	int rc, n_children = 0;
	char path[DR_PATH_MAX], *root_path;

	/* First, validate that the requested drc-name is not
	 * already owned by the partition.
	 */
	node = get_node_by_name(opts->usr_drc_name, PCI_NODES | VIO_NODES,
				&all_nodes);

	if (node) {
		if (node->is_owned) {
			err_msg("partition already owns %s\n",
				opts->usr_drc_name);
			free_node(all_nodes);
			return RC_ALREADY_OWN;
		}

		free_node(all_nodes);
	}

	rc = get_drc_by_name(opts->usr_drc_name, &drc, path, OFDT_BASE);
	if (rc)
		return rc;

	rc = acquire_drc(drc.index, &new_nodes);
	if (rc)
		return rc;

	/* We need the path starting at /pci or /vdevice, one or the other
	 * should exist.
	 */
	root_path = strstr(path, "/pci");
	if (root_path == NULL)
		root_path = strstr(path, "/vdevice");

	rc = add_nodes_to_kernel(root_path, new_nodes);
	if (rc) {
		dbg("add_nodes_to_kernel failed at %s\n", root_path);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");

		drc_recovery(drc.index);
		return -1;
	}

	/* Now that the node has been added to the device-tree, retrieve it.
	 * This also acts as a sanity check that everything up to this
	 * point has succeeded.
	 */
	node = get_node_by_name(drc.name, PCI_NODES | VIO_NODES, &all_nodes);
	if (node == NULL) {
		dbg("Could not get find \"%s\"\n", opts->usr_drc_name);
		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");

		drc_recovery(drc.index);
		return -1;
	}

	/* For PCI nodes */
	if (node->dev_type == PCI_DLPAR_DEV) {
		sprintf(path, "%s/%s", root_path, new_nodes->name);
		rc = acquire_hp_children(path, &n_children);
		if (rc) {
			err_msg("Unknown failure. Data may be out of sync and "
				"\nthe system may require a reboot.\n");

			drc_recovery(node->drc_index);
			goto slot_add_exit;
		}
	}

	/* Need to add node into sysfs which will "Hot plug" the node into
	 * pci world.
	 */
	rc = dlpar_io_kernel_op(DLPAR_ADD_SLOT, node->drc_name);
	if (rc) {
		char cmd[256];

		dbg("add %s into hotplug subsystem failed\n", node->drc_name);
		drc_recovery(node->drc_index);
		sprintf(cmd, "%s/%s", root_path, new_nodes->name);

		dbg("Removing newly added nodes with cmd: %s\n", cmd);
		rc = remove_nodes_from_kernel(cmd);
		if (rc)
			dbg("removing node %s failed\n", node->drc_name);
		else
			dbg("device node(s) for %s removed\n", node->drc_name);

		err_msg("Unknown failure. Data may be out of sync and \n"
			"the system may require a reboot.\n");
		goto slot_add_exit;
	}

	if (n_children) {
		rc = enable_hp_children(node->drc_name);
		if (rc) {
			err_msg("Configure adapter failed.\n");
			goto slot_add_exit;
		}

		dbg("adapter in node[%s] has been configured.\n",
		    node->drc_name);
	}

slot_add_exit:
	if (all_nodes)
		free_node(all_nodes);

	return rc;
}

int
valid_slot_options(struct options *opts)
{
	if (opts->usr_drc_name == NULL) {
		err_msg("A drc name must be specified\n");
		return -1;
	}

	return 0;
}

int
drslot_chrp_slot(struct options *opts)
{
	int rc;

	if (kernel_check(DLPAR_ADD_SLOT)) {
		err_msg("DLPAR slot operations are not supported on"
			"this kernel.");
		return -1;
	}

	switch (opts->action) {
	    case ADD:
		rc = add_slot(opts);
		break;

	    case REMOVE:
		rc = remove_slot(opts);
		break;

	    case QUERY:
		rc = query_slot(opts);
		break;

	    default:
		usage();
		rc = -1;
	}

	return rc;
}
